'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var querystring = require('querystring');
var Path = _interopDefault(require('path'));

// Make a map and return a function for checking if a key
// is in that map.
//
// IMPORTANT: all calls of this function must be prefixed with /*#__PURE__*/
// So that rollup can tree-shake them if necessary.
function makeMap(str, expectsLowerCase) {
    const map = Object.create(null);
    const list = str.split(',');
    for (let i = 0; i < list.length; i++) {
        map[list[i]] = true;
    }
    return expectsLowerCase ? val => !!map[val.toLowerCase()] : val => !!map[val];
}
/**
 * Always return false.
 */
const NO = () => false;
const extend = (a, b) => {
    for (const key in b) {
        a[key] = b[key];
    }
    return a;
};
const isArray = Array.isArray;
const cacheStringFunction = (fn) => {
    const cache = Object.create(null);
    return ((str) => {
        const hit = cache[str];
        return hit || (cache[str] = fn(str));
    });
};
const hyphenateRE = /\B([A-Z])/g;
const hyphenate = cacheStringFunction((str) => {
    return str.replace(hyphenateRE, '-$1').toLowerCase();
});

function defaultOnError(error) {
    throw error;
}
function createCompilerError(code, loc, messages, additionalMessage) {
    const msg =  code;
    const error = new SyntaxError(String(msg));
    error.code = code;
    error.loc = loc;
    return error;
}
const TELEPORT = Symbol( ``);
const SUSPENSE = Symbol( ``);
const KEEP_ALIVE = Symbol( ``);
const BASE_TRANSITION = Symbol( ``);

// AST Utilities ---------------------------------------------------------------
// Some expressions, e.g. sequence and conditional expressions, are never
// associated with template nodes, so their source locations are just a stub.
// Container types like CompoundExpression also don't need a real location.
const locStub = {
    source: '',
    start: { line: 1, column: 1, offset: 0 },
    end: { line: 1, column: 1, offset: 0 }
};
function createRoot(children, loc = locStub) {
    return {
        type: 0 /* ROOT */,
        children,
        helpers: [],
        components: [],
        directives: [],
        hoists: [],
        imports: [],
        cached: 0,
        temps: 0,
        codegenNode: undefined,
        loc
    };
}

const isBuiltInType = (tag, expected) => tag === expected || tag === hyphenate(expected);
function isCoreComponent(tag) {
    if (isBuiltInType(tag, 'Teleport')) {
        return TELEPORT;
    }
    else if (isBuiltInType(tag, 'Suspense')) {
        return SUSPENSE;
    }
    else if (isBuiltInType(tag, 'KeepAlive')) {
        return KEEP_ALIVE;
    }
    else if (isBuiltInType(tag, 'BaseTransition')) {
        return BASE_TRANSITION;
    }
}
function advancePositionWithClone(pos, source, numberOfCharacters = source.length) {
    return advancePositionWithMutation({ ...pos }, source, numberOfCharacters);
}
// advance by mutation without cloning (for performance reasons), since this
// gets called a lot in the parser
function advancePositionWithMutation(pos, source, numberOfCharacters = source.length) {
    let linesCount = 0;
    let lastNewLinePos = -1;
    for (let i = 0; i < numberOfCharacters; i++) {
        if (source.charCodeAt(i) === 10 /* newline char code */) {
            linesCount++;
            lastNewLinePos = i;
        }
    }
    pos.offset += numberOfCharacters;
    pos.line += linesCount;
    pos.column =
        lastNewLinePos === -1
            ? pos.column + numberOfCharacters
            : numberOfCharacters - lastNewLinePos;
    return pos;
}

// The default decoder only provides escapes for characters reserved as part of
// the tempalte syntax, and is only used if the custom renderer did not provide
// a platform-specific decoder.
const decodeRE = /&(gt|lt|amp|apos|quot);/g;
const decodeMap = {
    gt: '>',
    lt: '<',
    amp: '&',
    apos: "'",
    quot: '"'
};
const defaultParserOptions = {
    delimiters: [`{{`, `}}`],
    getNamespace: () => 0 /* HTML */,
    getTextMode: () => 0 /* DATA */,
    isVoidTag: NO,
    isPreTag: NO,
    isCustomElement: NO,
    decodeEntities: (rawText) => rawText.replace(decodeRE, (_, p1) => decodeMap[p1]),
    onError: defaultOnError
};
function baseParse(content, options = {}) {
    const context = createParserContext(content, options);
    const start = getCursor(context);
    return createRoot(parseChildren(context, 0 /* DATA */, []), getSelection(context, start));
}
function createParserContext(content, options) {
    return {
        options: {
            ...defaultParserOptions,
            ...options
        },
        column: 1,
        line: 1,
        offset: 0,
        originalSource: content,
        source: content,
        inPre: false,
        inVPre: false
    };
}
function parseChildren(context, mode, ancestors) {
    const parent = last(ancestors);
    const ns = parent ? parent.ns : 0 /* HTML */;
    const nodes = [];
    while (!isEnd(context, mode, ancestors)) {
        const s = context.source;
        let node = undefined;
        if (mode === 0 /* DATA */ || mode === 1 /* RCDATA */) {
            if (!context.inVPre && startsWith(s, context.options.delimiters[0])) {
                // '{{'
                node = parseInterpolation(context, mode);
            }
            else if (mode === 0 /* DATA */ && s[0] === '<') {
                // https://html.spec.whatwg.org/multipage/parsing.html#tag-open-state
                if (s.length === 1) {
                    emitError(context, 5 /* EOF_BEFORE_TAG_NAME */, 1);
                }
                else if (s[1] === '!') {
                    // https://html.spec.whatwg.org/multipage/parsing.html#markup-declaration-open-state
                    if (startsWith(s, '<!--')) {
                        node = parseComment(context);
                    }
                    else if (startsWith(s, '<!DOCTYPE')) {
                        // Ignore DOCTYPE by a limitation.
                        node = parseBogusComment(context);
                    }
                    else if (startsWith(s, '<![CDATA[')) {
                        if (ns !== 0 /* HTML */) {
                            node = parseCDATA(context, ancestors);
                        }
                        else {
                            emitError(context, 1 /* CDATA_IN_HTML_CONTENT */);
                            node = parseBogusComment(context);
                        }
                    }
                    else {
                        emitError(context, 11 /* INCORRECTLY_OPENED_COMMENT */);
                        node = parseBogusComment(context);
                    }
                }
                else if (s[1] === '/') {
                    // https://html.spec.whatwg.org/multipage/parsing.html#end-tag-open-state
                    if (s.length === 2) {
                        emitError(context, 5 /* EOF_BEFORE_TAG_NAME */, 2);
                    }
                    else if (s[2] === '>') {
                        emitError(context, 14 /* MISSING_END_TAG_NAME */, 2);
                        advanceBy(context, 3);
                        continue;
                    }
                    else if (/[a-z]/i.test(s[2])) {
                        emitError(context, 23 /* X_INVALID_END_TAG */);
                        parseTag(context, 1 /* End */, parent);
                        continue;
                    }
                    else {
                        emitError(context, 12 /* INVALID_FIRST_CHARACTER_OF_TAG_NAME */, 2);
                        node = parseBogusComment(context);
                    }
                }
                else if (/[a-z]/i.test(s[1])) {
                    node = parseElement(context, ancestors);
                }
                else if (s[1] === '?') {
                    emitError(context, 21 /* UNEXPECTED_QUESTION_MARK_INSTEAD_OF_TAG_NAME */, 1);
                    node = parseBogusComment(context);
                }
                else {
                    emitError(context, 12 /* INVALID_FIRST_CHARACTER_OF_TAG_NAME */, 1);
                }
            }
        }
        if (!node) {
            node = parseText(context, mode);
        }
        if (isArray(node)) {
            for (let i = 0; i < node.length; i++) {
                pushNode(nodes, node[i]);
            }
        }
        else {
            pushNode(nodes, node);
        }
    }
    // Whitespace management for more efficient output
    // (same as v2 whitespace: 'condense')
    let removedWhitespace = false;
    if (mode !== 2 /* RAWTEXT */) {
        if (!context.inPre) {
            for (let i = 0; i < nodes.length; i++) {
                const node = nodes[i];
                if (node.type === 2 /* TEXT */) {
                    if (!/[^\t\r\n\f ]/.test(node.content)) {
                        const prev = nodes[i - 1];
                        const next = nodes[i + 1];
                        // If:
                        // - the whitespace is the first or last node, or:
                        // - the whitespace is adjacent to a comment, or:
                        // - the whitespace is between two elements AND contains newline
                        // Then the whitespace is ignored.
                        if (!prev ||
                            !next ||
                            prev.type === 3 /* COMMENT */ ||
                            next.type === 3 /* COMMENT */ ||
                            (prev.type === 1 /* ELEMENT */ &&
                                next.type === 1 /* ELEMENT */ &&
                                /[\r\n]/.test(node.content))) {
                            removedWhitespace = true;
                            nodes[i] = null;
                        }
                        else {
                            // Otherwise, condensed consecutive whitespace inside the text down to
                            // a single space
                            node.content = ' ';
                        }
                    }
                    else {
                        node.content = node.content.replace(/[\t\r\n\f ]+/g, ' ');
                    }
                }
            }
        }
        else if (parent && context.options.isPreTag(parent.tag)) {
            // remove leading newline per html spec
            // https://html.spec.whatwg.org/multipage/grouping-content.html#the-pre-element
            const first = nodes[0];
            if (first && first.type === 2 /* TEXT */) {
                first.content = first.content.replace(/^\r?\n/, '');
            }
        }
    }
    return removedWhitespace ? nodes.filter(Boolean) : nodes;
}
function pushNode(nodes, node) {
    // ignore comments in production
    /* istanbul ignore next */
    if ( node.type === 3 /* COMMENT */) {
        return;
    }
    if (node.type === 2 /* TEXT */) {
        const prev = last(nodes);
        // Merge if both this and the previous node are text and those are
        // consecutive. This happens for cases like "a < b".
        if (prev &&
            prev.type === 2 /* TEXT */ &&
            prev.loc.end.offset === node.loc.start.offset) {
            prev.content += node.content;
            prev.loc.end = node.loc.end;
            prev.loc.source += node.loc.source;
            return;
        }
    }
    nodes.push(node);
}
function parseCDATA(context, ancestors) {
    advanceBy(context, 9);
    const nodes = parseChildren(context, 3 /* CDATA */, ancestors);
    if (context.source.length === 0) {
        emitError(context, 6 /* EOF_IN_CDATA */);
    }
    else {
        advanceBy(context, 3);
    }
    return nodes;
}
function parseComment(context) {
    const start = getCursor(context);
    let content;
    // Regular comment.
    const match = /--(\!)?>/.exec(context.source);
    if (!match) {
        content = context.source.slice(4);
        advanceBy(context, context.source.length);
        emitError(context, 7 /* EOF_IN_COMMENT */);
    }
    else {
        if (match.index <= 3) {
            emitError(context, 0 /* ABRUPT_CLOSING_OF_EMPTY_COMMENT */);
        }
        if (match[1]) {
            emitError(context, 10 /* INCORRECTLY_CLOSED_COMMENT */);
        }
        content = context.source.slice(4, match.index);
        // Advancing with reporting nested comments.
        const s = context.source.slice(0, match.index);
        let prevIndex = 1, nestedIndex = 0;
        while ((nestedIndex = s.indexOf('<!--', prevIndex)) !== -1) {
            advanceBy(context, nestedIndex - prevIndex + 1);
            if (nestedIndex + 4 < s.length) {
                emitError(context, 16 /* NESTED_COMMENT */);
            }
            prevIndex = nestedIndex + 1;
        }
        advanceBy(context, match.index + match[0].length - prevIndex + 1);
    }
    return {
        type: 3 /* COMMENT */,
        content,
        loc: getSelection(context, start)
    };
}
function parseBogusComment(context) {
    const start = getCursor(context);
    const contentStart = context.source[1] === '?' ? 1 : 2;
    let content;
    const closeIndex = context.source.indexOf('>');
    if (closeIndex === -1) {
        content = context.source.slice(contentStart);
        advanceBy(context, context.source.length);
    }
    else {
        content = context.source.slice(contentStart, closeIndex);
        advanceBy(context, closeIndex + 1);
    }
    return {
        type: 3 /* COMMENT */,
        content,
        loc: getSelection(context, start)
    };
}
function parseElement(context, ancestors) {
    // Start tag.
    const wasInPre = context.inPre;
    const wasInVPre = context.inVPre;
    const parent = last(ancestors);
    const element = parseTag(context, 0 /* Start */, parent);
    const isPreBoundary = context.inPre && !wasInPre;
    const isVPreBoundary = context.inVPre && !wasInVPre;
    if (element.isSelfClosing || context.options.isVoidTag(element.tag)) {
        return element;
    }
    // Children.
    ancestors.push(element);
    const mode = context.options.getTextMode(element.tag, element.ns, parent);
    const children = parseChildren(context, mode, ancestors);
    ancestors.pop();
    element.children = children;
    // End tag.
    if (startsWithEndTagOpen(context.source, element.tag)) {
        parseTag(context, 1 /* End */, parent);
    }
    else {
        emitError(context, 24 /* X_MISSING_END_TAG */, 0, element.loc.start);
        if (context.source.length === 0 && element.tag.toLowerCase() === 'script') {
            const first = children[0];
            if (first && startsWith(first.loc.source, '<!--')) {
                emitError(context, 8 /* EOF_IN_SCRIPT_HTML_COMMENT_LIKE_TEXT */);
            }
        }
    }
    element.loc = getSelection(context, element.loc.start);
    if (isPreBoundary) {
        context.inPre = false;
    }
    if (isVPreBoundary) {
        context.inVPre = false;
    }
    return element;
}
const isSpecialTemplateDirective = /*#__PURE__*/ makeMap(`if,else,else-if,for,slot`);
/**
 * Parse a tag (E.g. `<div id=a>`) with that type (start tag or end tag).
 */
function parseTag(context, type, parent) {
    // Tag open.
    const start = getCursor(context);
    const match = /^<\/?([a-z][^\t\r\n\f />]*)/i.exec(context.source);
    const tag = match[1];
    const ns = context.options.getNamespace(tag, parent);
    advanceBy(context, match[0].length);
    advanceSpaces(context);
    // save current state in case we need to re-parse attributes with v-pre
    const cursor = getCursor(context);
    const currentSource = context.source;
    // Attributes.
    let props = parseAttributes(context, type);
    // check <pre> tag
    if (context.options.isPreTag(tag)) {
        context.inPre = true;
    }
    // check v-pre
    if (!context.inVPre &&
        props.some(p => p.type === 7 /* DIRECTIVE */ && p.name === 'pre')) {
        context.inVPre = true;
        // reset context
        extend(context, cursor);
        context.source = currentSource;
        // re-parse attrs and filter out v-pre itself
        props = parseAttributes(context, type).filter(p => p.name !== 'v-pre');
    }
    // Tag close.
    let isSelfClosing = false;
    if (context.source.length === 0) {
        emitError(context, 9 /* EOF_IN_TAG */);
    }
    else {
        isSelfClosing = startsWith(context.source, '/>');
        if (type === 1 /* End */ && isSelfClosing) {
            emitError(context, 4 /* END_TAG_WITH_TRAILING_SOLIDUS */);
        }
        advanceBy(context, isSelfClosing ? 2 : 1);
    }
    let tagType = 0 /* ELEMENT */;
    const options = context.options;
    if (!context.inVPre && !options.isCustomElement(tag)) {
        const hasVIs = props.some(p => p.type === 7 /* DIRECTIVE */ && p.name === 'is');
        if (options.isNativeTag && !hasVIs) {
            if (!options.isNativeTag(tag))
                tagType = 1 /* COMPONENT */;
        }
        else if (hasVIs ||
            isCoreComponent(tag) ||
            (options.isBuiltInComponent && options.isBuiltInComponent(tag)) ||
            /^[A-Z]/.test(tag) ||
            tag === 'component') {
            tagType = 1 /* COMPONENT */;
        }
        if (tag === 'slot') {
            tagType = 2 /* SLOT */;
        }
        else if (tag === 'template' &&
            props.some(p => {
                return (p.type === 7 /* DIRECTIVE */ && isSpecialTemplateDirective(p.name));
            })) {
            tagType = 3 /* TEMPLATE */;
        }
    }
    return {
        type: 1 /* ELEMENT */,
        ns,
        tag,
        tagType,
        props,
        isSelfClosing,
        children: [],
        loc: getSelection(context, start),
        codegenNode: undefined // to be created during transform phase
    };
}
function parseAttributes(context, type) {
    const props = [];
    const attributeNames = new Set();
    while (context.source.length > 0 &&
        !startsWith(context.source, '>') &&
        !startsWith(context.source, '/>')) {
        if (startsWith(context.source, '/')) {
            emitError(context, 22 /* UNEXPECTED_SOLIDUS_IN_TAG */);
            advanceBy(context, 1);
            advanceSpaces(context);
            continue;
        }
        if (type === 1 /* End */) {
            emitError(context, 3 /* END_TAG_WITH_ATTRIBUTES */);
        }
        const attr = parseAttribute(context, attributeNames);
        if (type === 0 /* Start */) {
            props.push(attr);
        }
        if (/^[^\t\r\n\f />]/.test(context.source)) {
            emitError(context, 15 /* MISSING_WHITESPACE_BETWEEN_ATTRIBUTES */);
        }
        advanceSpaces(context);
    }
    return props;
}
function parseAttribute(context, nameSet) {
    // Name.
    const start = getCursor(context);
    const match = /^[^\t\r\n\f />][^\t\r\n\f />=]*/.exec(context.source);
    const name = match[0];
    if (nameSet.has(name)) {
        emitError(context, 2 /* DUPLICATE_ATTRIBUTE */);
    }
    nameSet.add(name);
    if (name[0] === '=') {
        emitError(context, 19 /* UNEXPECTED_EQUALS_SIGN_BEFORE_ATTRIBUTE_NAME */);
    }
    {
        const pattern = /["'<]/g;
        let m;
        while ((m = pattern.exec(name))) {
            emitError(context, 17 /* UNEXPECTED_CHARACTER_IN_ATTRIBUTE_NAME */, m.index);
        }
    }
    advanceBy(context, name.length);
    // Value
    let value = undefined;
    if (/^[\t\r\n\f ]*=/.test(context.source)) {
        advanceSpaces(context);
        advanceBy(context, 1);
        advanceSpaces(context);
        value = parseAttributeValue(context);
        if (!value) {
            emitError(context, 13 /* MISSING_ATTRIBUTE_VALUE */);
        }
    }
    const loc = getSelection(context, start);
    if (!context.inVPre && /^(v-|:|@|#)/.test(name)) {
        const match = /(?:^v-([a-z0-9-]+))?(?:(?::|^@|^#)([^\.]+))?(.+)?$/i.exec(name);
        let arg;
        if (match[2]) {
            const startOffset = name.indexOf(match[2]);
            const loc = getSelection(context, getNewPosition(context, start, startOffset), getNewPosition(context, start, startOffset + match[2].length));
            let content = match[2];
            let isStatic = true;
            if (content.startsWith('[')) {
                isStatic = false;
                if (!content.endsWith(']')) {
                    emitError(context, 26 /* X_MISSING_DYNAMIC_DIRECTIVE_ARGUMENT_END */);
                }
                content = content.substr(1, content.length - 2);
            }
            arg = {
                type: 4 /* SIMPLE_EXPRESSION */,
                content,
                isStatic,
                isConstant: isStatic,
                loc
            };
        }
        if (value && value.isQuoted) {
            const valueLoc = value.loc;
            valueLoc.start.offset++;
            valueLoc.start.column++;
            valueLoc.end = advancePositionWithClone(valueLoc.start, value.content);
            valueLoc.source = valueLoc.source.slice(1, -1);
        }
        return {
            type: 7 /* DIRECTIVE */,
            name: match[1] ||
                (startsWith(name, ':')
                    ? 'bind'
                    : startsWith(name, '@')
                        ? 'on'
                        : 'slot'),
            exp: value && {
                type: 4 /* SIMPLE_EXPRESSION */,
                content: value.content,
                isStatic: false,
                // Treat as non-constant by default. This can be potentially set to
                // true by `transformExpression` to make it eligible for hoisting.
                isConstant: false,
                loc: value.loc
            },
            arg,
            modifiers: match[3] ? match[3].substr(1).split('.') : [],
            loc
        };
    }
    return {
        type: 6 /* ATTRIBUTE */,
        name,
        value: value && {
            type: 2 /* TEXT */,
            content: value.content,
            loc: value.loc
        },
        loc
    };
}
function parseAttributeValue(context) {
    const start = getCursor(context);
    let content;
    const quote = context.source[0];
    const isQuoted = quote === `"` || quote === `'`;
    if (isQuoted) {
        // Quoted value.
        advanceBy(context, 1);
        const endIndex = context.source.indexOf(quote);
        if (endIndex === -1) {
            content = parseTextData(context, context.source.length, 4 /* ATTRIBUTE_VALUE */);
        }
        else {
            content = parseTextData(context, endIndex, 4 /* ATTRIBUTE_VALUE */);
            advanceBy(context, 1);
        }
    }
    else {
        // Unquoted
        const match = /^[^\t\r\n\f >]+/.exec(context.source);
        if (!match) {
            return undefined;
        }
        const unexpectedChars = /["'<=`]/g;
        let m;
        while ((m = unexpectedChars.exec(match[0]))) {
            emitError(context, 18 /* UNEXPECTED_CHARACTER_IN_UNQUOTED_ATTRIBUTE_VALUE */, m.index);
        }
        content = parseTextData(context, match[0].length, 4 /* ATTRIBUTE_VALUE */);
    }
    return { content, isQuoted, loc: getSelection(context, start) };
}
function parseInterpolation(context, mode) {
    const [open, close] = context.options.delimiters;
    const closeIndex = context.source.indexOf(close, open.length);
    if (closeIndex === -1) {
        emitError(context, 25 /* X_MISSING_INTERPOLATION_END */);
        return undefined;
    }
    const start = getCursor(context);
    advanceBy(context, open.length);
    const innerStart = getCursor(context);
    const innerEnd = getCursor(context);
    const rawContentLength = closeIndex - open.length;
    const rawContent = context.source.slice(0, rawContentLength);
    const preTrimContent = parseTextData(context, rawContentLength, mode);
    const content = preTrimContent.trim();
    const startOffset = preTrimContent.indexOf(content);
    if (startOffset > 0) {
        advancePositionWithMutation(innerStart, rawContent, startOffset);
    }
    const endOffset = rawContentLength - (preTrimContent.length - content.length - startOffset);
    advancePositionWithMutation(innerEnd, rawContent, endOffset);
    advanceBy(context, close.length);
    return {
        type: 5 /* INTERPOLATION */,
        content: {
            type: 4 /* SIMPLE_EXPRESSION */,
            isStatic: false,
            // Set `isConstant` to false by default and will decide in transformExpression
            isConstant: false,
            content,
            loc: getSelection(context, innerStart, innerEnd)
        },
        loc: getSelection(context, start)
    };
}
function parseText(context, mode) {
    const endTokens = ['<', context.options.delimiters[0]];
    if (mode === 3 /* CDATA */) {
        endTokens.push(']]>');
    }
    let endIndex = context.source.length;
    for (let i = 0; i < endTokens.length; i++) {
        const index = context.source.indexOf(endTokens[i], 1);
        if (index !== -1 && endIndex > index) {
            endIndex = index;
        }
    }
    const start = getCursor(context);
    const content = parseTextData(context, endIndex, mode);
    return {
        type: 2 /* TEXT */,
        content,
        loc: getSelection(context, start)
    };
}
/**
 * Get text data with a given length from the current location.
 * This translates HTML entities in the text data.
 */
function parseTextData(context, length, mode) {
    const rawText = context.source.slice(0, length);
    advanceBy(context, length);
    if (mode === 2 /* RAWTEXT */ ||
        mode === 3 /* CDATA */ ||
        rawText.indexOf('&') === -1) {
        return rawText;
    }
    else {
        // DATA or RCDATA containing "&"". Entity decoding required.
        return context.options.decodeEntities(rawText, mode === 4 /* ATTRIBUTE_VALUE */);
    }
}
function getCursor(context) {
    const { column, line, offset } = context;
    return { column, line, offset };
}
function getSelection(context, start, end) {
    end = end || getCursor(context);
    return {
        start,
        end,
        source: context.originalSource.slice(start.offset, end.offset)
    };
}
function last(xs) {
    return xs[xs.length - 1];
}
function startsWith(source, searchString) {
    return source.startsWith(searchString);
}
function advanceBy(context, numberOfCharacters) {
    const { source } = context;
    advancePositionWithMutation(context, source, numberOfCharacters);
    context.source = source.slice(numberOfCharacters);
}
function advanceSpaces(context) {
    const match = /^[\t\r\n\f ]+/.exec(context.source);
    if (match) {
        advanceBy(context, match[0].length);
    }
}
function getNewPosition(context, start, numberOfCharacters) {
    return advancePositionWithClone(start, context.originalSource.slice(start.offset, numberOfCharacters), numberOfCharacters);
}
function emitError(context, code, offset, loc = getCursor(context)) {
    if (offset) {
        loc.offset += offset;
        loc.column += offset;
    }
    context.options.onError(createCompilerError(code, {
        start: loc,
        end: loc,
        source: ''
    }));
}
function isEnd(context, mode, ancestors) {
    const s = context.source;
    switch (mode) {
        case 0 /* DATA */:
            if (startsWith(s, '</')) {
                //TODO: probably bad performance
                for (let i = ancestors.length - 1; i >= 0; --i) {
                    if (startsWithEndTagOpen(s, ancestors[i].tag)) {
                        return true;
                    }
                }
            }
            break;
        case 1 /* RCDATA */:
        case 2 /* RAWTEXT */: {
            const parent = last(ancestors);
            if (parent && startsWithEndTagOpen(s, parent.tag)) {
                return true;
            }
            break;
        }
        case 3 /* CDATA */:
            if (startsWith(s, ']]>')) {
                return true;
            }
            break;
    }
    return !s;
}
function startsWithEndTagOpen(source, tag) {
    return (startsWith(source, '</') &&
        source.substr(2, tag.length).toLowerCase() === tag.toLowerCase() &&
        /[\t\n\f />]/.test(source[2 + tag.length] || '>'));
}

/* -*- Mode: js; js-indent-level: 2; -*- */
/*
 * Copyright 2011 Mozilla Foundation and contributors
 * Licensed under the New BSD license. See LICENSE or:
 * http://opensource.org/licenses/BSD-3-Clause
 */

var intToCharMap = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/'.split('');

/**
 * Encode an integer in the range of 0 to 63 to a single base 64 digit.
 */
var encode = function (number) {
  if (0 <= number && number < intToCharMap.length) {
    return intToCharMap[number];
  }
  throw new TypeError("Must be between 0 and 63: " + number);
};

/**
 * Decode a single base 64 character code digit to an integer. Returns -1 on
 * failure.
 */
var decode = function (charCode) {
  var bigA = 65;     // 'A'
  var bigZ = 90;     // 'Z'

  var littleA = 97;  // 'a'
  var littleZ = 122; // 'z'

  var zero = 48;     // '0'
  var nine = 57;     // '9'

  var plus = 43;     // '+'
  var slash = 47;    // '/'

  var littleOffset = 26;
  var numberOffset = 52;

  // 0 - 25: ABCDEFGHIJKLMNOPQRSTUVWXYZ
  if (bigA <= charCode && charCode <= bigZ) {
    return (charCode - bigA);
  }

  // 26 - 51: abcdefghijklmnopqrstuvwxyz
  if (littleA <= charCode && charCode <= littleZ) {
    return (charCode - littleA + littleOffset);
  }

  // 52 - 61: 0123456789
  if (zero <= charCode && charCode <= nine) {
    return (charCode - zero + numberOffset);
  }

  // 62: +
  if (charCode == plus) {
    return 62;
  }

  // 63: /
  if (charCode == slash) {
    return 63;
  }

  // Invalid base64 digit.
  return -1;
};

var base64 = {
	encode: encode,
	decode: decode
};

/* -*- Mode: js; js-indent-level: 2; -*- */
/*
 * Copyright 2011 Mozilla Foundation and contributors
 * Licensed under the New BSD license. See LICENSE or:
 * http://opensource.org/licenses/BSD-3-Clause
 *
 * Based on the Base 64 VLQ implementation in Closure Compiler:
 * https://code.google.com/p/closure-compiler/source/browse/trunk/src/com/google/debugging/sourcemap/Base64VLQ.java
 *
 * Copyright 2011 The Closure Compiler Authors. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *  * Neither the name of Google Inc. nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */



// A single base 64 digit can contain 6 bits of data. For the base 64 variable
// length quantities we use in the source map spec, the first bit is the sign,
// the next four bits are the actual value, and the 6th bit is the
// continuation bit. The continuation bit tells us whether there are more
// digits in this value following this digit.
//
//   Continuation
//   |    Sign
//   |    |
//   V    V
//   101011

var VLQ_BASE_SHIFT = 5;

// binary: 100000
var VLQ_BASE = 1 << VLQ_BASE_SHIFT;

// binary: 011111
var VLQ_BASE_MASK = VLQ_BASE - 1;

// binary: 100000
var VLQ_CONTINUATION_BIT = VLQ_BASE;

/**
 * Converts from a two-complement value to a value where the sign bit is
 * placed in the least significant bit.  For example, as decimals:
 *   1 becomes 2 (10 binary), -1 becomes 3 (11 binary)
 *   2 becomes 4 (100 binary), -2 becomes 5 (101 binary)
 */
function toVLQSigned(aValue) {
  return aValue < 0
    ? ((-aValue) << 1) + 1
    : (aValue << 1) + 0;
}

/**
 * Converts to a two-complement value from a value where the sign bit is
 * placed in the least significant bit.  For example, as decimals:
 *   2 (10 binary) becomes 1, 3 (11 binary) becomes -1
 *   4 (100 binary) becomes 2, 5 (101 binary) becomes -2
 */
function fromVLQSigned(aValue) {
  var isNegative = (aValue & 1) === 1;
  var shifted = aValue >> 1;
  return isNegative
    ? -shifted
    : shifted;
}

/**
 * Returns the base 64 VLQ encoded value.
 */
var encode$1 = function base64VLQ_encode(aValue) {
  var encoded = "";
  var digit;

  var vlq = toVLQSigned(aValue);

  do {
    digit = vlq & VLQ_BASE_MASK;
    vlq >>>= VLQ_BASE_SHIFT;
    if (vlq > 0) {
      // There are still more digits in this value, so we must make sure the
      // continuation bit is marked.
      digit |= VLQ_CONTINUATION_BIT;
    }
    encoded += base64.encode(digit);
  } while (vlq > 0);

  return encoded;
};

/**
 * Decodes the next base 64 VLQ value from the given string and returns the
 * value and the rest of the string via the out parameter.
 */
var decode$1 = function base64VLQ_decode(aStr, aIndex, aOutParam) {
  var strLen = aStr.length;
  var result = 0;
  var shift = 0;
  var continuation, digit;

  do {
    if (aIndex >= strLen) {
      throw new Error("Expected more digits in base 64 VLQ value.");
    }

    digit = base64.decode(aStr.charCodeAt(aIndex++));
    if (digit === -1) {
      throw new Error("Invalid base64 digit: " + aStr.charAt(aIndex - 1));
    }

    continuation = !!(digit & VLQ_CONTINUATION_BIT);
    digit &= VLQ_BASE_MASK;
    result = result + (digit << shift);
    shift += VLQ_BASE_SHIFT;
  } while (continuation);

  aOutParam.value = fromVLQSigned(result);
  aOutParam.rest = aIndex;
};

var base64Vlq = {
	encode: encode$1,
	decode: decode$1
};

function createCommonjsModule(fn, module) {
	return module = { exports: {} }, fn(module, module.exports), module.exports;
}

var util = createCommonjsModule(function (module, exports) {
/* -*- Mode: js; js-indent-level: 2; -*- */
/*
 * Copyright 2011 Mozilla Foundation and contributors
 * Licensed under the New BSD license. See LICENSE or:
 * http://opensource.org/licenses/BSD-3-Clause
 */

/**
 * This is a helper function for getting values from parameter/options
 * objects.
 *
 * @param args The object we are extracting values from
 * @param name The name of the property we are getting.
 * @param defaultValue An optional value to return if the property is missing
 * from the object. If this is not specified and the property is missing, an
 * error will be thrown.
 */
function getArg(aArgs, aName, aDefaultValue) {
  if (aName in aArgs) {
    return aArgs[aName];
  } else if (arguments.length === 3) {
    return aDefaultValue;
  } else {
    throw new Error('"' + aName + '" is a required argument.');
  }
}
exports.getArg = getArg;

var urlRegexp = /^(?:([\w+\-.]+):)?\/\/(?:(\w+:\w+)@)?([\w.-]*)(?::(\d+))?(.*)$/;
var dataUrlRegexp = /^data:.+\,.+$/;

function urlParse(aUrl) {
  var match = aUrl.match(urlRegexp);
  if (!match) {
    return null;
  }
  return {
    scheme: match[1],
    auth: match[2],
    host: match[3],
    port: match[4],
    path: match[5]
  };
}
exports.urlParse = urlParse;

function urlGenerate(aParsedUrl) {
  var url = '';
  if (aParsedUrl.scheme) {
    url += aParsedUrl.scheme + ':';
  }
  url += '//';
  if (aParsedUrl.auth) {
    url += aParsedUrl.auth + '@';
  }
  if (aParsedUrl.host) {
    url += aParsedUrl.host;
  }
  if (aParsedUrl.port) {
    url += ":" + aParsedUrl.port;
  }
  if (aParsedUrl.path) {
    url += aParsedUrl.path;
  }
  return url;
}
exports.urlGenerate = urlGenerate;

/**
 * Normalizes a path, or the path portion of a URL:
 *
 * - Replaces consecutive slashes with one slash.
 * - Removes unnecessary '.' parts.
 * - Removes unnecessary '<dir>/..' parts.
 *
 * Based on code in the Node.js 'path' core module.
 *
 * @param aPath The path or url to normalize.
 */
function normalize(aPath) {
  var path = aPath;
  var url = urlParse(aPath);
  if (url) {
    if (!url.path) {
      return aPath;
    }
    path = url.path;
  }
  var isAbsolute = exports.isAbsolute(path);

  var parts = path.split(/\/+/);
  for (var part, up = 0, i = parts.length - 1; i >= 0; i--) {
    part = parts[i];
    if (part === '.') {
      parts.splice(i, 1);
    } else if (part === '..') {
      up++;
    } else if (up > 0) {
      if (part === '') {
        // The first part is blank if the path is absolute. Trying to go
        // above the root is a no-op. Therefore we can remove all '..' parts
        // directly after the root.
        parts.splice(i + 1, up);
        up = 0;
      } else {
        parts.splice(i, 2);
        up--;
      }
    }
  }
  path = parts.join('/');

  if (path === '') {
    path = isAbsolute ? '/' : '.';
  }

  if (url) {
    url.path = path;
    return urlGenerate(url);
  }
  return path;
}
exports.normalize = normalize;

/**
 * Joins two paths/URLs.
 *
 * @param aRoot The root path or URL.
 * @param aPath The path or URL to be joined with the root.
 *
 * - If aPath is a URL or a data URI, aPath is returned, unless aPath is a
 *   scheme-relative URL: Then the scheme of aRoot, if any, is prepended
 *   first.
 * - Otherwise aPath is a path. If aRoot is a URL, then its path portion
 *   is updated with the result and aRoot is returned. Otherwise the result
 *   is returned.
 *   - If aPath is absolute, the result is aPath.
 *   - Otherwise the two paths are joined with a slash.
 * - Joining for example 'http://' and 'www.example.com' is also supported.
 */
function join(aRoot, aPath) {
  if (aRoot === "") {
    aRoot = ".";
  }
  if (aPath === "") {
    aPath = ".";
  }
  var aPathUrl = urlParse(aPath);
  var aRootUrl = urlParse(aRoot);
  if (aRootUrl) {
    aRoot = aRootUrl.path || '/';
  }

  // `join(foo, '//www.example.org')`
  if (aPathUrl && !aPathUrl.scheme) {
    if (aRootUrl) {
      aPathUrl.scheme = aRootUrl.scheme;
    }
    return urlGenerate(aPathUrl);
  }

  if (aPathUrl || aPath.match(dataUrlRegexp)) {
    return aPath;
  }

  // `join('http://', 'www.example.com')`
  if (aRootUrl && !aRootUrl.host && !aRootUrl.path) {
    aRootUrl.host = aPath;
    return urlGenerate(aRootUrl);
  }

  var joined = aPath.charAt(0) === '/'
    ? aPath
    : normalize(aRoot.replace(/\/+$/, '') + '/' + aPath);

  if (aRootUrl) {
    aRootUrl.path = joined;
    return urlGenerate(aRootUrl);
  }
  return joined;
}
exports.join = join;

exports.isAbsolute = function (aPath) {
  return aPath.charAt(0) === '/' || urlRegexp.test(aPath);
};

/**
 * Make a path relative to a URL or another path.
 *
 * @param aRoot The root path or URL.
 * @param aPath The path or URL to be made relative to aRoot.
 */
function relative(aRoot, aPath) {
  if (aRoot === "") {
    aRoot = ".";
  }

  aRoot = aRoot.replace(/\/$/, '');

  // It is possible for the path to be above the root. In this case, simply
  // checking whether the root is a prefix of the path won't work. Instead, we
  // need to remove components from the root one by one, until either we find
  // a prefix that fits, or we run out of components to remove.
  var level = 0;
  while (aPath.indexOf(aRoot + '/') !== 0) {
    var index = aRoot.lastIndexOf("/");
    if (index < 0) {
      return aPath;
    }

    // If the only part of the root that is left is the scheme (i.e. http://,
    // file:///, etc.), one or more slashes (/), or simply nothing at all, we
    // have exhausted all components, so the path is not relative to the root.
    aRoot = aRoot.slice(0, index);
    if (aRoot.match(/^([^\/]+:\/)?\/*$/)) {
      return aPath;
    }

    ++level;
  }

  // Make sure we add a "../" for each component we removed from the root.
  return Array(level + 1).join("../") + aPath.substr(aRoot.length + 1);
}
exports.relative = relative;

var supportsNullProto = (function () {
  var obj = Object.create(null);
  return !('__proto__' in obj);
}());

function identity (s) {
  return s;
}

/**
 * Because behavior goes wacky when you set `__proto__` on objects, we
 * have to prefix all the strings in our set with an arbitrary character.
 *
 * See https://github.com/mozilla/source-map/pull/31 and
 * https://github.com/mozilla/source-map/issues/30
 *
 * @param String aStr
 */
function toSetString(aStr) {
  if (isProtoString(aStr)) {
    return '$' + aStr;
  }

  return aStr;
}
exports.toSetString = supportsNullProto ? identity : toSetString;

function fromSetString(aStr) {
  if (isProtoString(aStr)) {
    return aStr.slice(1);
  }

  return aStr;
}
exports.fromSetString = supportsNullProto ? identity : fromSetString;

function isProtoString(s) {
  if (!s) {
    return false;
  }

  var length = s.length;

  if (length < 9 /* "__proto__".length */) {
    return false;
  }

  if (s.charCodeAt(length - 1) !== 95  /* '_' */ ||
      s.charCodeAt(length - 2) !== 95  /* '_' */ ||
      s.charCodeAt(length - 3) !== 111 /* 'o' */ ||
      s.charCodeAt(length - 4) !== 116 /* 't' */ ||
      s.charCodeAt(length - 5) !== 111 /* 'o' */ ||
      s.charCodeAt(length - 6) !== 114 /* 'r' */ ||
      s.charCodeAt(length - 7) !== 112 /* 'p' */ ||
      s.charCodeAt(length - 8) !== 95  /* '_' */ ||
      s.charCodeAt(length - 9) !== 95  /* '_' */) {
    return false;
  }

  for (var i = length - 10; i >= 0; i--) {
    if (s.charCodeAt(i) !== 36 /* '$' */) {
      return false;
    }
  }

  return true;
}

/**
 * Comparator between two mappings where the original positions are compared.
 *
 * Optionally pass in `true` as `onlyCompareGenerated` to consider two
 * mappings with the same original source/line/column, but different generated
 * line and column the same. Useful when searching for a mapping with a
 * stubbed out mapping.
 */
function compareByOriginalPositions(mappingA, mappingB, onlyCompareOriginal) {
  var cmp = strcmp(mappingA.source, mappingB.source);
  if (cmp !== 0) {
    return cmp;
  }

  cmp = mappingA.originalLine - mappingB.originalLine;
  if (cmp !== 0) {
    return cmp;
  }

  cmp = mappingA.originalColumn - mappingB.originalColumn;
  if (cmp !== 0 || onlyCompareOriginal) {
    return cmp;
  }

  cmp = mappingA.generatedColumn - mappingB.generatedColumn;
  if (cmp !== 0) {
    return cmp;
  }

  cmp = mappingA.generatedLine - mappingB.generatedLine;
  if (cmp !== 0) {
    return cmp;
  }

  return strcmp(mappingA.name, mappingB.name);
}
exports.compareByOriginalPositions = compareByOriginalPositions;

/**
 * Comparator between two mappings with deflated source and name indices where
 * the generated positions are compared.
 *
 * Optionally pass in `true` as `onlyCompareGenerated` to consider two
 * mappings with the same generated line and column, but different
 * source/name/original line and column the same. Useful when searching for a
 * mapping with a stubbed out mapping.
 */
function compareByGeneratedPositionsDeflated(mappingA, mappingB, onlyCompareGenerated) {
  var cmp = mappingA.generatedLine - mappingB.generatedLine;
  if (cmp !== 0) {
    return cmp;
  }

  cmp = mappingA.generatedColumn - mappingB.generatedColumn;
  if (cmp !== 0 || onlyCompareGenerated) {
    return cmp;
  }

  cmp = strcmp(mappingA.source, mappingB.source);
  if (cmp !== 0) {
    return cmp;
  }

  cmp = mappingA.originalLine - mappingB.originalLine;
  if (cmp !== 0) {
    return cmp;
  }

  cmp = mappingA.originalColumn - mappingB.originalColumn;
  if (cmp !== 0) {
    return cmp;
  }

  return strcmp(mappingA.name, mappingB.name);
}
exports.compareByGeneratedPositionsDeflated = compareByGeneratedPositionsDeflated;

function strcmp(aStr1, aStr2) {
  if (aStr1 === aStr2) {
    return 0;
  }

  if (aStr1 === null) {
    return 1; // aStr2 !== null
  }

  if (aStr2 === null) {
    return -1; // aStr1 !== null
  }

  if (aStr1 > aStr2) {
    return 1;
  }

  return -1;
}

/**
 * Comparator between two mappings with inflated source and name strings where
 * the generated positions are compared.
 */
function compareByGeneratedPositionsInflated(mappingA, mappingB) {
  var cmp = mappingA.generatedLine - mappingB.generatedLine;
  if (cmp !== 0) {
    return cmp;
  }

  cmp = mappingA.generatedColumn - mappingB.generatedColumn;
  if (cmp !== 0) {
    return cmp;
  }

  cmp = strcmp(mappingA.source, mappingB.source);
  if (cmp !== 0) {
    return cmp;
  }

  cmp = mappingA.originalLine - mappingB.originalLine;
  if (cmp !== 0) {
    return cmp;
  }

  cmp = mappingA.originalColumn - mappingB.originalColumn;
  if (cmp !== 0) {
    return cmp;
  }

  return strcmp(mappingA.name, mappingB.name);
}
exports.compareByGeneratedPositionsInflated = compareByGeneratedPositionsInflated;

/**
 * Strip any JSON XSSI avoidance prefix from the string (as documented
 * in the source maps specification), and then parse the string as
 * JSON.
 */
function parseSourceMapInput(str) {
  return JSON.parse(str.replace(/^\)]}'[^\n]*\n/, ''));
}
exports.parseSourceMapInput = parseSourceMapInput;

/**
 * Compute the URL of a source given the the source root, the source's
 * URL, and the source map's URL.
 */
function computeSourceURL(sourceRoot, sourceURL, sourceMapURL) {
  sourceURL = sourceURL || '';

  if (sourceRoot) {
    // This follows what Chrome does.
    if (sourceRoot[sourceRoot.length - 1] !== '/' && sourceURL[0] !== '/') {
      sourceRoot += '/';
    }
    // The spec says:
    //   Line 4: An optional source root, useful for relocating source
    //   files on a server or removing repeated values in the
    //   “sources” entry.  This value is prepended to the individual
    //   entries in the “source” field.
    sourceURL = sourceRoot + sourceURL;
  }

  // Historically, SourceMapConsumer did not take the sourceMapURL as
  // a parameter.  This mode is still somewhat supported, which is why
  // this code block is conditional.  However, it's preferable to pass
  // the source map URL to SourceMapConsumer, so that this function
  // can implement the source URL resolution algorithm as outlined in
  // the spec.  This block is basically the equivalent of:
  //    new URL(sourceURL, sourceMapURL).toString()
  // ... except it avoids using URL, which wasn't available in the
  // older releases of node still supported by this library.
  //
  // The spec says:
  //   If the sources are not absolute URLs after prepending of the
  //   “sourceRoot”, the sources are resolved relative to the
  //   SourceMap (like resolving script src in a html document).
  if (sourceMapURL) {
    var parsed = urlParse(sourceMapURL);
    if (!parsed) {
      throw new Error("sourceMapURL could not be parsed");
    }
    if (parsed.path) {
      // Strip the last path component, but keep the "/".
      var index = parsed.path.lastIndexOf('/');
      if (index >= 0) {
        parsed.path = parsed.path.substring(0, index + 1);
      }
    }
    sourceURL = join(urlGenerate(parsed), sourceURL);
  }

  return normalize(sourceURL);
}
exports.computeSourceURL = computeSourceURL;
});

/* -*- Mode: js; js-indent-level: 2; -*- */
/*
 * Copyright 2011 Mozilla Foundation and contributors
 * Licensed under the New BSD license. See LICENSE or:
 * http://opensource.org/licenses/BSD-3-Clause
 */


var has = Object.prototype.hasOwnProperty;
var hasNativeMap = typeof Map !== "undefined";

/**
 * A data structure which is a combination of an array and a set. Adding a new
 * member is O(1), testing for membership is O(1), and finding the index of an
 * element is O(1). Removing elements from the set is not supported. Only
 * strings are supported for membership.
 */
function ArraySet() {
  this._array = [];
  this._set = hasNativeMap ? new Map() : Object.create(null);
}

/**
 * Static method for creating ArraySet instances from an existing array.
 */
ArraySet.fromArray = function ArraySet_fromArray(aArray, aAllowDuplicates) {
  var set = new ArraySet();
  for (var i = 0, len = aArray.length; i < len; i++) {
    set.add(aArray[i], aAllowDuplicates);
  }
  return set;
};

/**
 * Return how many unique items are in this ArraySet. If duplicates have been
 * added, than those do not count towards the size.
 *
 * @returns Number
 */
ArraySet.prototype.size = function ArraySet_size() {
  return hasNativeMap ? this._set.size : Object.getOwnPropertyNames(this._set).length;
};

/**
 * Add the given string to this set.
 *
 * @param String aStr
 */
ArraySet.prototype.add = function ArraySet_add(aStr, aAllowDuplicates) {
  var sStr = hasNativeMap ? aStr : util.toSetString(aStr);
  var isDuplicate = hasNativeMap ? this.has(aStr) : has.call(this._set, sStr);
  var idx = this._array.length;
  if (!isDuplicate || aAllowDuplicates) {
    this._array.push(aStr);
  }
  if (!isDuplicate) {
    if (hasNativeMap) {
      this._set.set(aStr, idx);
    } else {
      this._set[sStr] = idx;
    }
  }
};

/**
 * Is the given string a member of this set?
 *
 * @param String aStr
 */
ArraySet.prototype.has = function ArraySet_has(aStr) {
  if (hasNativeMap) {
    return this._set.has(aStr);
  } else {
    var sStr = util.toSetString(aStr);
    return has.call(this._set, sStr);
  }
};

/**
 * What is the index of the given string in the array?
 *
 * @param String aStr
 */
ArraySet.prototype.indexOf = function ArraySet_indexOf(aStr) {
  if (hasNativeMap) {
    var idx = this._set.get(aStr);
    if (idx >= 0) {
        return idx;
    }
  } else {
    var sStr = util.toSetString(aStr);
    if (has.call(this._set, sStr)) {
      return this._set[sStr];
    }
  }

  throw new Error('"' + aStr + '" is not in the set.');
};

/**
 * What is the element at the given index?
 *
 * @param Number aIdx
 */
ArraySet.prototype.at = function ArraySet_at(aIdx) {
  if (aIdx >= 0 && aIdx < this._array.length) {
    return this._array[aIdx];
  }
  throw new Error('No element indexed by ' + aIdx);
};

/**
 * Returns the array representation of this set (which has the proper indices
 * indicated by indexOf). Note that this is a copy of the internal array used
 * for storing the members so that no one can mess with internal state.
 */
ArraySet.prototype.toArray = function ArraySet_toArray() {
  return this._array.slice();
};

var ArraySet_1 = ArraySet;

var arraySet = {
	ArraySet: ArraySet_1
};

/* -*- Mode: js; js-indent-level: 2; -*- */
/*
 * Copyright 2014 Mozilla Foundation and contributors
 * Licensed under the New BSD license. See LICENSE or:
 * http://opensource.org/licenses/BSD-3-Clause
 */



/**
 * Determine whether mappingB is after mappingA with respect to generated
 * position.
 */
function generatedPositionAfter(mappingA, mappingB) {
  // Optimized for most common case
  var lineA = mappingA.generatedLine;
  var lineB = mappingB.generatedLine;
  var columnA = mappingA.generatedColumn;
  var columnB = mappingB.generatedColumn;
  return lineB > lineA || lineB == lineA && columnB >= columnA ||
         util.compareByGeneratedPositionsInflated(mappingA, mappingB) <= 0;
}

/**
 * A data structure to provide a sorted view of accumulated mappings in a
 * performance conscious manner. It trades a neglibable overhead in general
 * case for a large speedup in case of mappings being added in order.
 */
function MappingList() {
  this._array = [];
  this._sorted = true;
  // Serves as infimum
  this._last = {generatedLine: -1, generatedColumn: 0};
}

/**
 * Iterate through internal items. This method takes the same arguments that
 * `Array.prototype.forEach` takes.
 *
 * NOTE: The order of the mappings is NOT guaranteed.
 */
MappingList.prototype.unsortedForEach =
  function MappingList_forEach(aCallback, aThisArg) {
    this._array.forEach(aCallback, aThisArg);
  };

/**
 * Add the given source mapping.
 *
 * @param Object aMapping
 */
MappingList.prototype.add = function MappingList_add(aMapping) {
  if (generatedPositionAfter(this._last, aMapping)) {
    this._last = aMapping;
    this._array.push(aMapping);
  } else {
    this._sorted = false;
    this._array.push(aMapping);
  }
};

/**
 * Returns the flat, sorted array of mappings. The mappings are sorted by
 * generated position.
 *
 * WARNING: This method returns internal data without copying, for
 * performance. The return value must NOT be mutated, and should be treated as
 * an immutable borrow. If you want to take ownership, you must make your own
 * copy.
 */
MappingList.prototype.toArray = function MappingList_toArray() {
  if (!this._sorted) {
    this._array.sort(util.compareByGeneratedPositionsInflated);
    this._sorted = true;
  }
  return this._array;
};

var MappingList_1 = MappingList;

var mappingList = {
	MappingList: MappingList_1
};

/* -*- Mode: js; js-indent-level: 2; -*- */
/*
 * Copyright 2011 Mozilla Foundation and contributors
 * Licensed under the New BSD license. See LICENSE or:
 * http://opensource.org/licenses/BSD-3-Clause
 */



var ArraySet$1 = arraySet.ArraySet;
var MappingList$1 = mappingList.MappingList;

/**
 * An instance of the SourceMapGenerator represents a source map which is
 * being built incrementally. You may pass an object with the following
 * properties:
 *
 *   - file: The filename of the generated source.
 *   - sourceRoot: A root for all relative URLs in this source map.
 */
function SourceMapGenerator(aArgs) {
  if (!aArgs) {
    aArgs = {};
  }
  this._file = util.getArg(aArgs, 'file', null);
  this._sourceRoot = util.getArg(aArgs, 'sourceRoot', null);
  this._skipValidation = util.getArg(aArgs, 'skipValidation', false);
  this._sources = new ArraySet$1();
  this._names = new ArraySet$1();
  this._mappings = new MappingList$1();
  this._sourcesContents = null;
}

SourceMapGenerator.prototype._version = 3;

/**
 * Creates a new SourceMapGenerator based on a SourceMapConsumer
 *
 * @param aSourceMapConsumer The SourceMap.
 */
SourceMapGenerator.fromSourceMap =
  function SourceMapGenerator_fromSourceMap(aSourceMapConsumer) {
    var sourceRoot = aSourceMapConsumer.sourceRoot;
    var generator = new SourceMapGenerator({
      file: aSourceMapConsumer.file,
      sourceRoot: sourceRoot
    });
    aSourceMapConsumer.eachMapping(function (mapping) {
      var newMapping = {
        generated: {
          line: mapping.generatedLine,
          column: mapping.generatedColumn
        }
      };

      if (mapping.source != null) {
        newMapping.source = mapping.source;
        if (sourceRoot != null) {
          newMapping.source = util.relative(sourceRoot, newMapping.source);
        }

        newMapping.original = {
          line: mapping.originalLine,
          column: mapping.originalColumn
        };

        if (mapping.name != null) {
          newMapping.name = mapping.name;
        }
      }

      generator.addMapping(newMapping);
    });
    aSourceMapConsumer.sources.forEach(function (sourceFile) {
      var sourceRelative = sourceFile;
      if (sourceRoot !== null) {
        sourceRelative = util.relative(sourceRoot, sourceFile);
      }

      if (!generator._sources.has(sourceRelative)) {
        generator._sources.add(sourceRelative);
      }

      var content = aSourceMapConsumer.sourceContentFor(sourceFile);
      if (content != null) {
        generator.setSourceContent(sourceFile, content);
      }
    });
    return generator;
  };

/**
 * Add a single mapping from original source line and column to the generated
 * source's line and column for this source map being created. The mapping
 * object should have the following properties:
 *
 *   - generated: An object with the generated line and column positions.
 *   - original: An object with the original line and column positions.
 *   - source: The original source file (relative to the sourceRoot).
 *   - name: An optional original token name for this mapping.
 */
SourceMapGenerator.prototype.addMapping =
  function SourceMapGenerator_addMapping(aArgs) {
    var generated = util.getArg(aArgs, 'generated');
    var original = util.getArg(aArgs, 'original', null);
    var source = util.getArg(aArgs, 'source', null);
    var name = util.getArg(aArgs, 'name', null);

    if (!this._skipValidation) {
      this._validateMapping(generated, original, source, name);
    }

    if (source != null) {
      source = String(source);
      if (!this._sources.has(source)) {
        this._sources.add(source);
      }
    }

    if (name != null) {
      name = String(name);
      if (!this._names.has(name)) {
        this._names.add(name);
      }
    }

    this._mappings.add({
      generatedLine: generated.line,
      generatedColumn: generated.column,
      originalLine: original != null && original.line,
      originalColumn: original != null && original.column,
      source: source,
      name: name
    });
  };

/**
 * Set the source content for a source file.
 */
SourceMapGenerator.prototype.setSourceContent =
  function SourceMapGenerator_setSourceContent(aSourceFile, aSourceContent) {
    var source = aSourceFile;
    if (this._sourceRoot != null) {
      source = util.relative(this._sourceRoot, source);
    }

    if (aSourceContent != null) {
      // Add the source content to the _sourcesContents map.
      // Create a new _sourcesContents map if the property is null.
      if (!this._sourcesContents) {
        this._sourcesContents = Object.create(null);
      }
      this._sourcesContents[util.toSetString(source)] = aSourceContent;
    } else if (this._sourcesContents) {
      // Remove the source file from the _sourcesContents map.
      // If the _sourcesContents map is empty, set the property to null.
      delete this._sourcesContents[util.toSetString(source)];
      if (Object.keys(this._sourcesContents).length === 0) {
        this._sourcesContents = null;
      }
    }
  };

/**
 * Applies the mappings of a sub-source-map for a specific source file to the
 * source map being generated. Each mapping to the supplied source file is
 * rewritten using the supplied source map. Note: The resolution for the
 * resulting mappings is the minimium of this map and the supplied map.
 *
 * @param aSourceMapConsumer The source map to be applied.
 * @param aSourceFile Optional. The filename of the source file.
 *        If omitted, SourceMapConsumer's file property will be used.
 * @param aSourceMapPath Optional. The dirname of the path to the source map
 *        to be applied. If relative, it is relative to the SourceMapConsumer.
 *        This parameter is needed when the two source maps aren't in the same
 *        directory, and the source map to be applied contains relative source
 *        paths. If so, those relative source paths need to be rewritten
 *        relative to the SourceMapGenerator.
 */
SourceMapGenerator.prototype.applySourceMap =
  function SourceMapGenerator_applySourceMap(aSourceMapConsumer, aSourceFile, aSourceMapPath) {
    var sourceFile = aSourceFile;
    // If aSourceFile is omitted, we will use the file property of the SourceMap
    if (aSourceFile == null) {
      if (aSourceMapConsumer.file == null) {
        throw new Error(
          'SourceMapGenerator.prototype.applySourceMap requires either an explicit source file, ' +
          'or the source map\'s "file" property. Both were omitted.'
        );
      }
      sourceFile = aSourceMapConsumer.file;
    }
    var sourceRoot = this._sourceRoot;
    // Make "sourceFile" relative if an absolute Url is passed.
    if (sourceRoot != null) {
      sourceFile = util.relative(sourceRoot, sourceFile);
    }
    // Applying the SourceMap can add and remove items from the sources and
    // the names array.
    var newSources = new ArraySet$1();
    var newNames = new ArraySet$1();

    // Find mappings for the "sourceFile"
    this._mappings.unsortedForEach(function (mapping) {
      if (mapping.source === sourceFile && mapping.originalLine != null) {
        // Check if it can be mapped by the source map, then update the mapping.
        var original = aSourceMapConsumer.originalPositionFor({
          line: mapping.originalLine,
          column: mapping.originalColumn
        });
        if (original.source != null) {
          // Copy mapping
          mapping.source = original.source;
          if (aSourceMapPath != null) {
            mapping.source = util.join(aSourceMapPath, mapping.source);
          }
          if (sourceRoot != null) {
            mapping.source = util.relative(sourceRoot, mapping.source);
          }
          mapping.originalLine = original.line;
          mapping.originalColumn = original.column;
          if (original.name != null) {
            mapping.name = original.name;
          }
        }
      }

      var source = mapping.source;
      if (source != null && !newSources.has(source)) {
        newSources.add(source);
      }

      var name = mapping.name;
      if (name != null && !newNames.has(name)) {
        newNames.add(name);
      }

    }, this);
    this._sources = newSources;
    this._names = newNames;

    // Copy sourcesContents of applied map.
    aSourceMapConsumer.sources.forEach(function (sourceFile) {
      var content = aSourceMapConsumer.sourceContentFor(sourceFile);
      if (content != null) {
        if (aSourceMapPath != null) {
          sourceFile = util.join(aSourceMapPath, sourceFile);
        }
        if (sourceRoot != null) {
          sourceFile = util.relative(sourceRoot, sourceFile);
        }
        this.setSourceContent(sourceFile, content);
      }
    }, this);
  };

/**
 * A mapping can have one of the three levels of data:
 *
 *   1. Just the generated position.
 *   2. The Generated position, original position, and original source.
 *   3. Generated and original position, original source, as well as a name
 *      token.
 *
 * To maintain consistency, we validate that any new mapping being added falls
 * in to one of these categories.
 */
SourceMapGenerator.prototype._validateMapping =
  function SourceMapGenerator_validateMapping(aGenerated, aOriginal, aSource,
                                              aName) {
    // When aOriginal is truthy but has empty values for .line and .column,
    // it is most likely a programmer error. In this case we throw a very
    // specific error message to try to guide them the right way.
    // For example: https://github.com/Polymer/polymer-bundler/pull/519
    if (aOriginal && typeof aOriginal.line !== 'number' && typeof aOriginal.column !== 'number') {
        throw new Error(
            'original.line and original.column are not numbers -- you probably meant to omit ' +
            'the original mapping entirely and only map the generated position. If so, pass ' +
            'null for the original mapping instead of an object with empty or null values.'
        );
    }

    if (aGenerated && 'line' in aGenerated && 'column' in aGenerated
        && aGenerated.line > 0 && aGenerated.column >= 0
        && !aOriginal && !aSource && !aName) {
      // Case 1.
      return;
    }
    else if (aGenerated && 'line' in aGenerated && 'column' in aGenerated
             && aOriginal && 'line' in aOriginal && 'column' in aOriginal
             && aGenerated.line > 0 && aGenerated.column >= 0
             && aOriginal.line > 0 && aOriginal.column >= 0
             && aSource) {
      // Cases 2 and 3.
      return;
    }
    else {
      throw new Error('Invalid mapping: ' + JSON.stringify({
        generated: aGenerated,
        source: aSource,
        original: aOriginal,
        name: aName
      }));
    }
  };

/**
 * Serialize the accumulated mappings in to the stream of base 64 VLQs
 * specified by the source map format.
 */
SourceMapGenerator.prototype._serializeMappings =
  function SourceMapGenerator_serializeMappings() {
    var previousGeneratedColumn = 0;
    var previousGeneratedLine = 1;
    var previousOriginalColumn = 0;
    var previousOriginalLine = 0;
    var previousName = 0;
    var previousSource = 0;
    var result = '';
    var next;
    var mapping;
    var nameIdx;
    var sourceIdx;

    var mappings = this._mappings.toArray();
    for (var i = 0, len = mappings.length; i < len; i++) {
      mapping = mappings[i];
      next = '';

      if (mapping.generatedLine !== previousGeneratedLine) {
        previousGeneratedColumn = 0;
        while (mapping.generatedLine !== previousGeneratedLine) {
          next += ';';
          previousGeneratedLine++;
        }
      }
      else {
        if (i > 0) {
          if (!util.compareByGeneratedPositionsInflated(mapping, mappings[i - 1])) {
            continue;
          }
          next += ',';
        }
      }

      next += base64Vlq.encode(mapping.generatedColumn
                                 - previousGeneratedColumn);
      previousGeneratedColumn = mapping.generatedColumn;

      if (mapping.source != null) {
        sourceIdx = this._sources.indexOf(mapping.source);
        next += base64Vlq.encode(sourceIdx - previousSource);
        previousSource = sourceIdx;

        // lines are stored 0-based in SourceMap spec version 3
        next += base64Vlq.encode(mapping.originalLine - 1
                                   - previousOriginalLine);
        previousOriginalLine = mapping.originalLine - 1;

        next += base64Vlq.encode(mapping.originalColumn
                                   - previousOriginalColumn);
        previousOriginalColumn = mapping.originalColumn;

        if (mapping.name != null) {
          nameIdx = this._names.indexOf(mapping.name);
          next += base64Vlq.encode(nameIdx - previousName);
          previousName = nameIdx;
        }
      }

      result += next;
    }

    return result;
  };

SourceMapGenerator.prototype._generateSourcesContent =
  function SourceMapGenerator_generateSourcesContent(aSources, aSourceRoot) {
    return aSources.map(function (source) {
      if (!this._sourcesContents) {
        return null;
      }
      if (aSourceRoot != null) {
        source = util.relative(aSourceRoot, source);
      }
      var key = util.toSetString(source);
      return Object.prototype.hasOwnProperty.call(this._sourcesContents, key)
        ? this._sourcesContents[key]
        : null;
    }, this);
  };

/**
 * Externalize the source map.
 */
SourceMapGenerator.prototype.toJSON =
  function SourceMapGenerator_toJSON() {
    var map = {
      version: this._version,
      sources: this._sources.toArray(),
      names: this._names.toArray(),
      mappings: this._serializeMappings()
    };
    if (this._file != null) {
      map.file = this._file;
    }
    if (this._sourceRoot != null) {
      map.sourceRoot = this._sourceRoot;
    }
    if (this._sourcesContents) {
      map.sourcesContent = this._generateSourcesContent(map.sources, map.sourceRoot);
    }

    return map;
  };

/**
 * Render the source map being generated to a string.
 */
SourceMapGenerator.prototype.toString =
  function SourceMapGenerator_toString() {
    return JSON.stringify(this.toJSON());
  };

var SourceMapGenerator_1 = SourceMapGenerator;

var sourceMapGenerator = {
	SourceMapGenerator: SourceMapGenerator_1
};

/*
 * Copyright 2009-2011 Mozilla Foundation and contributors
 * Licensed under the New BSD license. See LICENSE.txt or:
 * http://opensource.org/licenses/BSD-3-Clause
 */
var SourceMapGenerator$1 = sourceMapGenerator.SourceMapGenerator;

const range = 2;
function generateCodeFrame(source, start = 0, end = source.length) {
    const lines = source.split(/\r?\n/);
    let count = 0;
    const res = [];
    for (let i = 0; i < lines.length; i++) {
        count += lines[i].length + 1;
        if (count >= start) {
            for (let j = i - range; j <= i + range || end > count; j++) {
                if (j < 0 || j >= lines.length)
                    continue;
                const line = j + 1;
                res.push(`${line}${' '.repeat(3 - String(line).length)}|  ${lines[j]}`);
                const lineLength = lines[j].length;
                if (j === i) {
                    // push underline
                    const pad = start - (count - lineLength) + 1;
                    const length = Math.max(1, end > count ? lineLength - pad : end - start);
                    res.push(`   |  ` + ' '.repeat(pad) + '^'.repeat(length));
                }
                else if (j > i) {
                    if (end > count) {
                        const length = Math.max(Math.min(end - count, lineLength), 1);
                        res.push(`   |  ` + '^'.repeat(length));
                    }
                    count += lineLength + 1;
                }
            }
            break;
        }
    }
    return res.join('\n');
}

function parse(source, { sourceMap = true, filename = 'component.vue', sourceRoot = '', pad = false, shouldPadTemplate = false, } = {}) {
    const descriptor = {
        filename,
        template: null,
        script: null,
        styles: [],
        customBlocks: [],
    };
    const errors = [];
    const ast = baseParse(source, {
        // there are no components at SFC parsing level
        isNativeTag: () => true,
        // preserve all whitespaces
        isPreTag: () => true,
        getTextMode: (tag, _ns, parent) => {
            // all top level elements except <template> are parsed as raw text
            // containers
            if (!parent && tag !== 'template') {
                return 2 /* RAWTEXT */;
            }
            else {
                return 0 /* DATA */;
            }
        },
        onError: (e) => {
            errors.push(e);
        },
    });
    ast.children.forEach((node) => {
        if (node.type !== 1 /* ELEMENT */) {
            return;
        }
        if (!node.children.length && !hasSrc(node)) {
            return;
        }
        switch (node.tag) {
            case 'template':
                if (!descriptor.template) {
                    descriptor.template = createBlock(node, source, shouldPadTemplate ? pad : false);
                }
                else {
                    warnDuplicateBlock(source, filename, node);
                }
                break;
            case 'script':
                if (!descriptor.script) {
                    descriptor.script = createBlock(node, source, pad);
                }
                else {
                    warnDuplicateBlock(source, filename, node);
                }
                break;
            case 'style':
                descriptor.styles.push(createBlock(node, source, pad));
                break;
            default:
                descriptor.customBlocks.push(createBlock(node, source, pad));
                break;
        }
    });
    if (sourceMap) {
        const genMap = (block) => {
            if (block && !block.src) {
                block.map = generateSourceMap(filename, source, block.content, sourceRoot, pad && (block.type !== 'template' || shouldPadTemplate)
                    ? block.loc.start.line - 1
                    : 0);
            }
        };
        genMap(descriptor.template);
        genMap(descriptor.script);
        descriptor.styles.forEach(genMap);
        descriptor.customBlocks.forEach(genMap);
    }
    const result = {
        descriptor,
        errors,
    };
    return result;
}
function warnDuplicateBlock(source, filename, node) {
    const codeFrame = generateCodeFrame(source, node.loc.start.offset, node.loc.end.offset);
    const location = `${filename}:${node.loc.start.line}:${node.loc.start.column}`;
    console.warn(`Single file component can contain only one ${node.tag} element (${location}):\n\n${codeFrame}`);
}
function createBlock(node, source, pad) {
    const type = node.tag;
    let { start, end } = node.loc;
    let content = '';
    if (node.children.length) {
        start = node.children[0].loc.start;
        end = node.children[node.children.length - 1].loc.end;
        content = source.slice(start.offset, end.offset);
    }
    const loc = {
        source: content,
        start,
        end,
    };
    const attrs = {};
    const block = {
        type,
        content,
        loc,
        attrs,
    };
    if (pad) {
        block.content = padContent(source, block, pad) + block.content;
    }
    node.props.forEach((p) => {
        if (p.type === 6 /* ATTRIBUTE */) {
            attrs[p.name] = p.value ? p.value.content || true : true;
            if (p.name === 'lang') {
                block.lang = p.value && p.value.content;
            }
            else if (p.name === 'src') {
                block.src = p.value && p.value.content;
            }
            else if (type === 'style') {
                if (p.name === 'scoped') {
                    block.scoped = true;
                }
                else if (p.name === 'module') {
                    block.module = attrs[p.name];
                }
            }
            else if (type === 'template' && p.name === 'functional') {
                block.functional = true;
            }
        }
    });
    return block;
}
const splitRE = /\r?\n/g;
const emptyRE = /^(?:\/\/)?\s*$/;
const replaceRE = /./g;
function generateSourceMap(filename, source, generated, sourceRoot, lineOffset) {
    const map = new SourceMapGenerator$1({
        file: filename.replace(/\\/g, '/'),
        sourceRoot: sourceRoot.replace(/\\/g, '/'),
    });
    map.setSourceContent(filename, source);
    generated.split(splitRE).forEach((line, index) => {
        if (!emptyRE.test(line)) {
            map.addMapping({
                source: filename,
                original: {
                    line: index + 1 + lineOffset,
                    column: 0,
                },
                generated: {
                    line: index + 1,
                    column: 0,
                },
            });
        }
    });
    return JSON.parse(map.toString());
}
function padContent(content, block, pad) {
    content = content.slice(0, block.loc.start.offset);
    if (pad === 'space') {
        return content.replace(replaceRE, ' ');
    }
    else {
        const offset = content.split(splitRE).length;
        const padChar = block.type === 'script' && !block.lang ? '//\n' : '\n';
        return Array(offset).join(padChar);
    }
}
function hasSrc(node) {
    return node.props.some((p) => {
        if (p.type !== 6 /* ATTRIBUTE */) {
            return false;
        }
        return p.name === 'src';
    });
}

/* --------------------------------------------------------------------------------------------
 * Copyright (c) Microsoft Corporation. All rights reserved.
 * Licensed under the MIT License. See License.txt in the project root for license information.
 * ------------------------------------------------------------------------------------------ */
var FullTextDocument = /** @class */ (function () {
    function FullTextDocument(uri, languageId, version, content) {
        this._uri = uri;
        this._languageId = languageId;
        this._version = version;
        this._content = content;
        this._lineOffsets = undefined;
    }
    Object.defineProperty(FullTextDocument.prototype, "uri", {
        get: function () {
            return this._uri;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(FullTextDocument.prototype, "languageId", {
        get: function () {
            return this._languageId;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(FullTextDocument.prototype, "version", {
        get: function () {
            return this._version;
        },
        enumerable: true,
        configurable: true
    });
    FullTextDocument.prototype.getText = function (range) {
        if (range) {
            var start = this.offsetAt(range.start);
            var end = this.offsetAt(range.end);
            return this._content.substring(start, end);
        }
        return this._content;
    };
    FullTextDocument.prototype.update = function (changes, version) {
        for (var _i = 0, changes_1 = changes; _i < changes_1.length; _i++) {
            var change = changes_1[_i];
            if (FullTextDocument.isIncremental(change)) {
                // makes sure start is before end
                var range = getWellformedRange(change.range);
                // update content
                var startOffset = this.offsetAt(range.start);
                var endOffset = this.offsetAt(range.end);
                this._content = this._content.substring(0, startOffset) + change.text + this._content.substring(endOffset, this._content.length);
                // update the offsets
                var startLine = Math.max(range.start.line, 0);
                var endLine = Math.max(range.end.line, 0);
                var lineOffsets = this._lineOffsets;
                var addedLineOffsets = computeLineOffsets(change.text, false, startOffset);
                if (endLine - startLine === addedLineOffsets.length) {
                    for (var i = 0, len = addedLineOffsets.length; i < len; i++) {
                        lineOffsets[i + startLine + 1] = addedLineOffsets[i];
                    }
                }
                else {
                    if (addedLineOffsets.length < 10000) {
                        lineOffsets.splice.apply(lineOffsets, [startLine + 1, endLine - startLine].concat(addedLineOffsets));
                    }
                    else { // avoid too many arguments for splice
                        this._lineOffsets = lineOffsets = lineOffsets.slice(0, startLine + 1).concat(addedLineOffsets, lineOffsets.slice(endLine + 1));
                    }
                }
                var diff = change.text.length - (endOffset - startOffset);
                if (diff !== 0) {
                    for (var i = startLine + 1 + addedLineOffsets.length, len = lineOffsets.length; i < len; i++) {
                        lineOffsets[i] = lineOffsets[i] + diff;
                    }
                }
            }
            else if (FullTextDocument.isFull(change)) {
                this._content = change.text;
                this._lineOffsets = undefined;
            }
            else {
                throw new Error('Unknown change event received');
            }
        }
        this._version = version;
    };
    FullTextDocument.prototype.getLineOffsets = function () {
        if (this._lineOffsets === undefined) {
            this._lineOffsets = computeLineOffsets(this._content, true);
        }
        return this._lineOffsets;
    };
    FullTextDocument.prototype.positionAt = function (offset) {
        offset = Math.max(Math.min(offset, this._content.length), 0);
        var lineOffsets = this.getLineOffsets();
        var low = 0, high = lineOffsets.length;
        if (high === 0) {
            return { line: 0, character: offset };
        }
        while (low < high) {
            var mid = Math.floor((low + high) / 2);
            if (lineOffsets[mid] > offset) {
                high = mid;
            }
            else {
                low = mid + 1;
            }
        }
        // low is the least x for which the line offset is larger than the current offset
        // or array.length if no line offset is larger than the current offset
        var line = low - 1;
        return { line: line, character: offset - lineOffsets[line] };
    };
    FullTextDocument.prototype.offsetAt = function (position) {
        var lineOffsets = this.getLineOffsets();
        if (position.line >= lineOffsets.length) {
            return this._content.length;
        }
        else if (position.line < 0) {
            return 0;
        }
        var lineOffset = lineOffsets[position.line];
        var nextLineOffset = (position.line + 1 < lineOffsets.length) ? lineOffsets[position.line + 1] : this._content.length;
        return Math.max(Math.min(lineOffset + position.character, nextLineOffset), lineOffset);
    };
    Object.defineProperty(FullTextDocument.prototype, "lineCount", {
        get: function () {
            return this.getLineOffsets().length;
        },
        enumerable: true,
        configurable: true
    });
    FullTextDocument.isIncremental = function (event) {
        var candidate = event;
        return candidate !== undefined && candidate !== null &&
            typeof candidate.text === 'string' && candidate.range !== undefined &&
            (candidate.rangeLength === undefined || typeof candidate.rangeLength === 'number');
    };
    FullTextDocument.isFull = function (event) {
        var candidate = event;
        return candidate !== undefined && candidate !== null &&
            typeof candidate.text === 'string' && candidate.range === undefined && candidate.rangeLength === undefined;
    };
    return FullTextDocument;
}());
var TextDocument;
(function (TextDocument) {
    /**
     * Creates a new text document.
     *
     * @param uri The document's uri.
     * @param languageId  The document's language Id.
     * @param version The document's initial version number.
     * @param content The document's content.
     */
    function create(uri, languageId, version, content) {
        return new FullTextDocument(uri, languageId, version, content);
    }
    TextDocument.create = create;
    /**
     * Updates a TextDocument by modifing its content.
     *
     * @param document the document to update. Only documents created by TextDocument.create are valid inputs.
     * @param changes the changes to apply to the document.
     * @returns The updated TextDocument. Note: That's the same document instance passed in as first parameter.
     *
     */
    function update(document, changes, version) {
        if (document instanceof FullTextDocument) {
            document.update(changes, version);
            return document;
        }
        else {
            throw new Error('TextDocument.update: document must be created by TextDocument.create');
        }
    }
    TextDocument.update = update;
    function applyEdits(document, edits) {
        var text = document.getText();
        var sortedEdits = mergeSort(edits.map(getWellformedEdit), function (a, b) {
            var diff = a.range.start.line - b.range.start.line;
            if (diff === 0) {
                return a.range.start.character - b.range.start.character;
            }
            return diff;
        });
        var lastModifiedOffset = 0;
        var spans = [];
        for (var _i = 0, sortedEdits_1 = sortedEdits; _i < sortedEdits_1.length; _i++) {
            var e = sortedEdits_1[_i];
            var startOffset = document.offsetAt(e.range.start);
            if (startOffset < lastModifiedOffset) {
                throw new Error('Overlapping edit');
            }
            else if (startOffset > lastModifiedOffset) {
                spans.push(text.substring(lastModifiedOffset, startOffset));
            }
            if (e.newText.length) {
                spans.push(e.newText);
            }
            lastModifiedOffset = document.offsetAt(e.range.end);
        }
        spans.push(text.substr(lastModifiedOffset));
        return spans.join('');
    }
    TextDocument.applyEdits = applyEdits;
})(TextDocument || (TextDocument = {}));
function mergeSort(data, compare) {
    if (data.length <= 1) {
        // sorted
        return data;
    }
    var p = (data.length / 2) | 0;
    var left = data.slice(0, p);
    var right = data.slice(p);
    mergeSort(left, compare);
    mergeSort(right, compare);
    var leftIdx = 0;
    var rightIdx = 0;
    var i = 0;
    while (leftIdx < left.length && rightIdx < right.length) {
        var ret = compare(left[leftIdx], right[rightIdx]);
        if (ret <= 0) {
            // smaller_equal -> take left to preserve order
            data[i++] = left[leftIdx++];
        }
        else {
            // greater -> take right
            data[i++] = right[rightIdx++];
        }
    }
    while (leftIdx < left.length) {
        data[i++] = left[leftIdx++];
    }
    while (rightIdx < right.length) {
        data[i++] = right[rightIdx++];
    }
    return data;
}
function computeLineOffsets(text, isAtLineStart, textOffset) {
    if (textOffset === void 0) { textOffset = 0; }
    var result = isAtLineStart ? [textOffset] : [];
    for (var i = 0; i < text.length; i++) {
        var ch = text.charCodeAt(i);
        if (ch === 13 /* CarriageReturn */ || ch === 10 /* LineFeed */) {
            if (ch === 13 /* CarriageReturn */ && i + 1 < text.length && text.charCodeAt(i + 1) === 10 /* LineFeed */) {
                i++;
            }
            result.push(textOffset + i + 1);
        }
    }
    return result;
}
function getWellformedRange(range) {
    var start = range.start;
    var end = range.end;
    if (start.line > end.line || (start.line === end.line && start.character > end.character)) {
        return { start: end, end: start };
    }
    return range;
}
function getWellformedEdit(textEdit) {
    var range = getWellformedRange(textEdit.range);
    if (range !== textEdit.range) {
        return { newText: textEdit.newText, range: range };
    }
    return textEdit;
}

/*---------------------------------------------------------------------------------------------
 *  Copyright (c) Microsoft Corporation. All rights reserved.
 *  Licensed under the MIT License. See License.txt in the project root for license information.
 *--------------------------------------------------------------------------------------------*/
var __extends = (null && null.__extends) || (function () {
    var extendStatics = function (d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
        return extendStatics(d, b);
    };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
var _a;
var isWindows;
if (typeof process === 'object') {
    isWindows = process.platform === 'win32';
}
else if (typeof navigator === 'object') {
    var userAgent = navigator.userAgent;
    isWindows = userAgent.indexOf('Windows') >= 0;
}
//#endregion
var _schemePattern = /^\w[\w\d+.-]*$/;
var _singleSlashStart = /^\//;
var _doubleSlashStart = /^\/\//;
function _validateUri(ret, _strict) {
    // scheme, must be set
    if (!ret.scheme && _strict) {
        throw new Error("[UriError]: Scheme is missing: {scheme: \"\", authority: \"" + ret.authority + "\", path: \"" + ret.path + "\", query: \"" + ret.query + "\", fragment: \"" + ret.fragment + "\"}");
    }
    // scheme, https://tools.ietf.org/html/rfc3986#section-3.1
    // ALPHA *( ALPHA / DIGIT / "+" / "-" / "." )
    if (ret.scheme && !_schemePattern.test(ret.scheme)) {
        throw new Error('[UriError]: Scheme contains illegal characters.');
    }
    // path, http://tools.ietf.org/html/rfc3986#section-3.3
    // If a URI contains an authority component, then the path component
    // must either be empty or begin with a slash ("/") character.  If a URI
    // does not contain an authority component, then the path cannot begin
    // with two slash characters ("//").
    if (ret.path) {
        if (ret.authority) {
            if (!_singleSlashStart.test(ret.path)) {
                throw new Error('[UriError]: If a URI contains an authority component, then the path component must either be empty or begin with a slash ("/") character');
            }
        }
        else {
            if (_doubleSlashStart.test(ret.path)) {
                throw new Error('[UriError]: If a URI does not contain an authority component, then the path cannot begin with two slash characters ("//")');
            }
        }
    }
}
// for a while we allowed uris *without* schemes and this is the migration
// for them, e.g. an uri without scheme and without strict-mode warns and falls
// back to the file-scheme. that should cause the least carnage and still be a
// clear warning
function _schemeFix(scheme, _strict) {
    if (!scheme && !_strict) {
        return 'file';
    }
    return scheme;
}
// implements a bit of https://tools.ietf.org/html/rfc3986#section-5
function _referenceResolution(scheme, path) {
    // the slash-character is our 'default base' as we don't
    // support constructing URIs relative to other URIs. This
    // also means that we alter and potentially break paths.
    // see https://tools.ietf.org/html/rfc3986#section-5.1.4
    switch (scheme) {
        case 'https':
        case 'http':
        case 'file':
            if (!path) {
                path = _slash;
            }
            else if (path[0] !== _slash) {
                path = _slash + path;
            }
            break;
    }
    return path;
}
var _empty = '';
var _slash = '/';
var _regexp = /^(([^:/?#]+?):)?(\/\/([^/?#]*))?([^?#]*)(\?([^#]*))?(#(.*))?/;
/**
 * Uniform Resource Identifier (URI) http://tools.ietf.org/html/rfc3986.
 * This class is a simple parser which creates the basic component parts
 * (http://tools.ietf.org/html/rfc3986#section-3) with minimal validation
 * and encoding.
 *
 *       foo://example.com:8042/over/there?name=ferret#nose
 *       \_/   \______________/\_________/ \_________/ \__/
 *        |           |            |            |        |
 *     scheme     authority       path        query   fragment
 *        |   _____________________|__
 *       / \ /                        \
 *       urn:example:animal:ferret:nose
 */
var URI = /** @class */ (function () {
    /**
     * @internal
     */
    function URI(schemeOrData, authority, path, query, fragment, _strict) {
        if (_strict === void 0) { _strict = false; }
        if (typeof schemeOrData === 'object') {
            this.scheme = schemeOrData.scheme || _empty;
            this.authority = schemeOrData.authority || _empty;
            this.path = schemeOrData.path || _empty;
            this.query = schemeOrData.query || _empty;
            this.fragment = schemeOrData.fragment || _empty;
            // no validation because it's this URI
            // that creates uri components.
            // _validateUri(this);
        }
        else {
            this.scheme = _schemeFix(schemeOrData, _strict);
            this.authority = authority || _empty;
            this.path = _referenceResolution(this.scheme, path || _empty);
            this.query = query || _empty;
            this.fragment = fragment || _empty;
            _validateUri(this, _strict);
        }
    }
    URI.isUri = function (thing) {
        if (thing instanceof URI) {
            return true;
        }
        if (!thing) {
            return false;
        }
        return typeof thing.authority === 'string'
            && typeof thing.fragment === 'string'
            && typeof thing.path === 'string'
            && typeof thing.query === 'string'
            && typeof thing.scheme === 'string'
            && typeof thing.fsPath === 'function'
            && typeof thing.with === 'function'
            && typeof thing.toString === 'function';
    };
    Object.defineProperty(URI.prototype, "fsPath", {
        // ---- filesystem path -----------------------
        /**
         * Returns a string representing the corresponding file system path of this URI.
         * Will handle UNC paths, normalizes windows drive letters to lower-case, and uses the
         * platform specific path separator.
         *
         * * Will *not* validate the path for invalid characters and semantics.
         * * Will *not* look at the scheme of this URI.
         * * The result shall *not* be used for display purposes but for accessing a file on disk.
         *
         *
         * The *difference* to `URI#path` is the use of the platform specific separator and the handling
         * of UNC paths. See the below sample of a file-uri with an authority (UNC path).
         *
         * ```ts
            const u = URI.parse('file://server/c$/folder/file.txt')
            u.authority === 'server'
            u.path === '/shares/c$/file.txt'
            u.fsPath === '\\server\c$\folder\file.txt'
        ```
         *
         * Using `URI#path` to read a file (using fs-apis) would not be enough because parts of the path,
         * namely the server name, would be missing. Therefore `URI#fsPath` exists - it's sugar to ease working
         * with URIs that represent files on disk (`file` scheme).
         */
        get: function () {
            // if (this.scheme !== 'file') {
            // 	console.warn(`[UriError] calling fsPath with scheme ${this.scheme}`);
            // }
            return _makeFsPath(this);
        },
        enumerable: true,
        configurable: true
    });
    // ---- modify to new -------------------------
    URI.prototype.with = function (change) {
        if (!change) {
            return this;
        }
        var scheme = change.scheme, authority = change.authority, path = change.path, query = change.query, fragment = change.fragment;
        if (scheme === undefined) {
            scheme = this.scheme;
        }
        else if (scheme === null) {
            scheme = _empty;
        }
        if (authority === undefined) {
            authority = this.authority;
        }
        else if (authority === null) {
            authority = _empty;
        }
        if (path === undefined) {
            path = this.path;
        }
        else if (path === null) {
            path = _empty;
        }
        if (query === undefined) {
            query = this.query;
        }
        else if (query === null) {
            query = _empty;
        }
        if (fragment === undefined) {
            fragment = this.fragment;
        }
        else if (fragment === null) {
            fragment = _empty;
        }
        if (scheme === this.scheme
            && authority === this.authority
            && path === this.path
            && query === this.query
            && fragment === this.fragment) {
            return this;
        }
        return new _URI(scheme, authority, path, query, fragment);
    };
    // ---- parse & validate ------------------------
    /**
     * Creates a new URI from a string, e.g. `http://www.msft.com/some/path`,
     * `file:///usr/home`, or `scheme:with/path`.
     *
     * @param value A string which represents an URI (see `URI#toString`).
     */
    URI.parse = function (value, _strict) {
        if (_strict === void 0) { _strict = false; }
        var match = _regexp.exec(value);
        if (!match) {
            return new _URI(_empty, _empty, _empty, _empty, _empty);
        }
        return new _URI(match[2] || _empty, decodeURIComponent(match[4] || _empty), decodeURIComponent(match[5] || _empty), decodeURIComponent(match[7] || _empty), decodeURIComponent(match[9] || _empty), _strict);
    };
    /**
     * Creates a new URI from a file system path, e.g. `c:\my\files`,
     * `/usr/home`, or `\\server\share\some\path`.
     *
     * The *difference* between `URI#parse` and `URI#file` is that the latter treats the argument
     * as path, not as stringified-uri. E.g. `URI.file(path)` is **not the same as**
     * `URI.parse('file://' + path)` because the path might contain characters that are
     * interpreted (# and ?). See the following sample:
     * ```ts
    const good = URI.file('/coding/c#/project1');
    good.scheme === 'file';
    good.path === '/coding/c#/project1';
    good.fragment === '';
    const bad = URI.parse('file://' + '/coding/c#/project1');
    bad.scheme === 'file';
    bad.path === '/coding/c'; // path is now broken
    bad.fragment === '/project1';
    ```
     *
     * @param path A file system path (see `URI#fsPath`)
     */
    URI.file = function (path) {
        var authority = _empty;
        // normalize to fwd-slashes on windows,
        // on other systems bwd-slashes are valid
        // filename character, eg /f\oo/ba\r.txt
        if (isWindows) {
            path = path.replace(/\\/g, _slash);
        }
        // check for authority as used in UNC shares
        // or use the path as given
        if (path[0] === _slash && path[1] === _slash) {
            var idx = path.indexOf(_slash, 2);
            if (idx === -1) {
                authority = path.substring(2);
                path = _slash;
            }
            else {
                authority = path.substring(2, idx);
                path = path.substring(idx) || _slash;
            }
        }
        return new _URI('file', authority, path, _empty, _empty);
    };
    URI.from = function (components) {
        return new _URI(components.scheme, components.authority, components.path, components.query, components.fragment);
    };
    // ---- printing/externalize ---------------------------
    /**
     * Creates a string representation for this URI. It's guaranteed that calling
     * `URI.parse` with the result of this function creates an URI which is equal
     * to this URI.
     *
     * * The result shall *not* be used for display purposes but for externalization or transport.
     * * The result will be encoded using the percentage encoding and encoding happens mostly
     * ignore the scheme-specific encoding rules.
     *
     * @param skipEncoding Do not encode the result, default is `false`
     */
    URI.prototype.toString = function (skipEncoding) {
        if (skipEncoding === void 0) { skipEncoding = false; }
        return _asFormatted(this, skipEncoding);
    };
    URI.prototype.toJSON = function () {
        return this;
    };
    URI.revive = function (data) {
        if (!data) {
            return data;
        }
        else if (data instanceof URI) {
            return data;
        }
        else {
            var result = new _URI(data);
            result._formatted = data.external;
            result._fsPath = data._sep === _pathSepMarker ? data.fsPath : null;
            return result;
        }
    };
    return URI;
}());
var _pathSepMarker = isWindows ? 1 : undefined;
// tslint:disable-next-line:class-name
var _URI = /** @class */ (function (_super) {
    __extends(_URI, _super);
    function _URI() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        _this._formatted = null;
        _this._fsPath = null;
        return _this;
    }
    Object.defineProperty(_URI.prototype, "fsPath", {
        get: function () {
            if (!this._fsPath) {
                this._fsPath = _makeFsPath(this);
            }
            return this._fsPath;
        },
        enumerable: true,
        configurable: true
    });
    _URI.prototype.toString = function (skipEncoding) {
        if (skipEncoding === void 0) { skipEncoding = false; }
        if (!skipEncoding) {
            if (!this._formatted) {
                this._formatted = _asFormatted(this, false);
            }
            return this._formatted;
        }
        else {
            // we don't cache that
            return _asFormatted(this, true);
        }
    };
    _URI.prototype.toJSON = function () {
        var res = {
            $mid: 1
        };
        // cached state
        if (this._fsPath) {
            res.fsPath = this._fsPath;
            res._sep = _pathSepMarker;
        }
        if (this._formatted) {
            res.external = this._formatted;
        }
        // uri components
        if (this.path) {
            res.path = this.path;
        }
        if (this.scheme) {
            res.scheme = this.scheme;
        }
        if (this.authority) {
            res.authority = this.authority;
        }
        if (this.query) {
            res.query = this.query;
        }
        if (this.fragment) {
            res.fragment = this.fragment;
        }
        return res;
    };
    return _URI;
}(URI));
// reserved characters: https://tools.ietf.org/html/rfc3986#section-2.2
var encodeTable = (_a = {},
    _a[58 /* Colon */] = '%3A',
    _a[47 /* Slash */] = '%2F',
    _a[63 /* QuestionMark */] = '%3F',
    _a[35 /* Hash */] = '%23',
    _a[91 /* OpenSquareBracket */] = '%5B',
    _a[93 /* CloseSquareBracket */] = '%5D',
    _a[64 /* AtSign */] = '%40',
    _a[33 /* ExclamationMark */] = '%21',
    _a[36 /* DollarSign */] = '%24',
    _a[38 /* Ampersand */] = '%26',
    _a[39 /* SingleQuote */] = '%27',
    _a[40 /* OpenParen */] = '%28',
    _a[41 /* CloseParen */] = '%29',
    _a[42 /* Asterisk */] = '%2A',
    _a[43 /* Plus */] = '%2B',
    _a[44 /* Comma */] = '%2C',
    _a[59 /* Semicolon */] = '%3B',
    _a[61 /* Equals */] = '%3D',
    _a[32 /* Space */] = '%20',
    _a);
function encodeURIComponentFast(uriComponent, allowSlash) {
    var res = undefined;
    var nativeEncodePos = -1;
    for (var pos = 0; pos < uriComponent.length; pos++) {
        var code = uriComponent.charCodeAt(pos);
        // unreserved characters: https://tools.ietf.org/html/rfc3986#section-2.3
        if ((code >= 97 /* a */ && code <= 122 /* z */)
            || (code >= 65 /* A */ && code <= 90 /* Z */)
            || (code >= 48 /* Digit0 */ && code <= 57 /* Digit9 */)
            || code === 45 /* Dash */
            || code === 46 /* Period */
            || code === 95 /* Underline */
            || code === 126 /* Tilde */
            || (allowSlash && code === 47 /* Slash */)) {
            // check if we are delaying native encode
            if (nativeEncodePos !== -1) {
                res += encodeURIComponent(uriComponent.substring(nativeEncodePos, pos));
                nativeEncodePos = -1;
            }
            // check if we write into a new string (by default we try to return the param)
            if (res !== undefined) {
                res += uriComponent.charAt(pos);
            }
        }
        else {
            // encoding needed, we need to allocate a new string
            if (res === undefined) {
                res = uriComponent.substr(0, pos);
            }
            // check with default table first
            var escaped = encodeTable[code];
            if (escaped !== undefined) {
                // check if we are delaying native encode
                if (nativeEncodePos !== -1) {
                    res += encodeURIComponent(uriComponent.substring(nativeEncodePos, pos));
                    nativeEncodePos = -1;
                }
                // append escaped variant to result
                res += escaped;
            }
            else if (nativeEncodePos === -1) {
                // use native encode only when needed
                nativeEncodePos = pos;
            }
        }
    }
    if (nativeEncodePos !== -1) {
        res += encodeURIComponent(uriComponent.substring(nativeEncodePos));
    }
    return res !== undefined ? res : uriComponent;
}
function encodeURIComponentMinimal(path) {
    var res = undefined;
    for (var pos = 0; pos < path.length; pos++) {
        var code = path.charCodeAt(pos);
        if (code === 35 /* Hash */ || code === 63 /* QuestionMark */) {
            if (res === undefined) {
                res = path.substr(0, pos);
            }
            res += encodeTable[code];
        }
        else {
            if (res !== undefined) {
                res += path[pos];
            }
        }
    }
    return res !== undefined ? res : path;
}
/**
 * Compute `fsPath` for the given uri
 */
function _makeFsPath(uri) {
    var value;
    if (uri.authority && uri.path.length > 1 && uri.scheme === 'file') {
        // unc path: file://shares/c$/far/boo
        value = "//" + uri.authority + uri.path;
    }
    else if (uri.path.charCodeAt(0) === 47 /* Slash */
        && (uri.path.charCodeAt(1) >= 65 /* A */ && uri.path.charCodeAt(1) <= 90 /* Z */ || uri.path.charCodeAt(1) >= 97 /* a */ && uri.path.charCodeAt(1) <= 122 /* z */)
        && uri.path.charCodeAt(2) === 58 /* Colon */) {
        // windows drive letter: file:///c:/far/boo
        value = uri.path[1].toLowerCase() + uri.path.substr(2);
    }
    else {
        // other path
        value = uri.path;
    }
    if (isWindows) {
        value = value.replace(/\//g, '\\');
    }
    return value;
}
/**
 * Create the external version of a uri
 */
function _asFormatted(uri, skipEncoding) {
    var encoder = !skipEncoding
        ? encodeURIComponentFast
        : encodeURIComponentMinimal;
    var res = '';
    var scheme = uri.scheme, authority = uri.authority, path = uri.path, query = uri.query, fragment = uri.fragment;
    if (scheme) {
        res += scheme;
        res += ':';
    }
    if (authority || scheme === 'file') {
        res += _slash;
        res += _slash;
    }
    if (authority) {
        var idx = authority.indexOf('@');
        if (idx !== -1) {
            // <user>@<auth>
            var userinfo = authority.substr(0, idx);
            authority = authority.substr(idx + 1);
            idx = userinfo.indexOf(':');
            if (idx === -1) {
                res += encoder(userinfo, false);
            }
            else {
                // <user>:<pass>@<auth>
                res += encoder(userinfo.substr(0, idx), false);
                res += ':';
                res += encoder(userinfo.substr(idx + 1), false);
            }
            res += '@';
        }
        authority = authority.toLowerCase();
        idx = authority.indexOf(':');
        if (idx === -1) {
            res += encoder(authority, false);
        }
        else {
            // <auth>:<port>
            res += encoder(authority.substr(0, idx), false);
            res += authority.substr(idx);
        }
    }
    if (path) {
        // lower-case windows drive letters in /C:/fff or C:/fff
        if (path.length >= 3 && path.charCodeAt(0) === 47 /* Slash */ && path.charCodeAt(2) === 58 /* Colon */) {
            var code = path.charCodeAt(1);
            if (code >= 65 /* A */ && code <= 90 /* Z */) {
                path = "/" + String.fromCharCode(code + 32) + ":" + path.substr(3); // "/c:".length === 3
            }
        }
        else if (path.length >= 2 && path.charCodeAt(1) === 58 /* Colon */) {
            var code = path.charCodeAt(0);
            if (code >= 65 /* A */ && code <= 90 /* Z */) {
                path = String.fromCharCode(code + 32) + ":" + path.substr(2); // "/c:".length === 3
            }
        }
        // encode the rest of the path
        res += encoder(path, true);
    }
    if (query) {
        res += '?';
        res += encoder(query, false);
    }
    if (fragment) {
        res += '#';
        res += !skipEncoding ? encodeURIComponentFast(fragment, false) : fragment;
    }
    return res;
}

// Vue File: foo.vue
// Virtual Script File: foo.vue____script.js
// Virtual Template File: foo.vue____template.vue-html
// Virtual Render File: foo.vue____render.js
// Virtual Style File: foo.vue____style__0.css
const virtualFileNameSep = '____';
function isVueFile(fileName) {
    return fileName.endsWith('.vue');
}
function isVirtualFile(fileName) {
    return fileName.indexOf('.vue' + virtualFileNameSep) > 0;
}
function asUri(fileNameOrUri) {
    if (/^[a-z-]+:\/\//i.test(fileNameOrUri))
        return fileNameOrUri;
    return VueTextDocument.fsPathToUri(fileNameOrUri);
}
class VirtualTextDocument {
    constructor(internal, container, selector) {
        this.internal = internal;
        this.container = container;
        this.selector = selector;
        this.uriObject = URI.parse(this.internal.uri);
    }
    get version() {
        return this.internal.version;
    }
    get languageId() {
        return this.internal.languageId;
    }
    get lineCount() {
        return this.internal.lineCount;
    }
    get uri() {
        return this.internal.uri;
    }
    get fsPath() {
        return this.uriObject.fsPath;
    }
    get content() {
        return this.getText();
    }
    getText(range) {
        return this.internal.getText(range);
    }
    positionAt(offset) {
        return this.internal.positionAt(offset);
    }
    offsetAt(position) {
        return this.internal.offsetAt(position);
    }
    toJSON() {
        return {
            uri: this.uri,
            languageId: this.languageId,
            vertion: this.version,
            content: this.getText(),
        };
    }
    static create(container, uri, languageId, version, content, selector) {
        return new VirtualTextDocument(TextDocument.create(uri, languageId, version, content), container, selector);
    }
    static update(document, changes, version) {
        document.internal = TextDocument.update(document.internal, changes, version);
        return document;
    }
}
const BlockLanguageMap = {
    script: 'js',
    template: 'vue-html',
    style: 'css',
    docs: 'md',
};
class VueTextDocument {
    constructor(internal) {
        this.internal = internal;
        this._blocks = [];
        this._parseErrors = [];
        this.documents = {};
        this.uriObject = URI.parse(this.internal.uri);
        this.parse();
    }
    get version() {
        return this.internal.version;
    }
    get languageId() {
        return this.internal.languageId;
    }
    get lineCount() {
        return this.internal.lineCount;
    }
    get uri() {
        return this.internal.uri;
    }
    get fsPath() {
        return this.uriObject.fsPath;
    }
    get descriptor() {
        return this._descriptor;
    }
    get blocks() {
        return this._blocks;
    }
    get parseErrors() {
        return this._parseErrors;
    }
    get content() {
        return this.getText();
    }
    getText(range) {
        return this.internal.getText(range);
    }
    positionAt(offset) {
        return this.internal.positionAt(offset);
    }
    offsetAt(position) {
        return this.internal.offsetAt(position);
    }
    blockAt(positionOrOffset) {
        const offset = typeof positionOrOffset === 'number'
            ? positionOrOffset
            : this.offsetAt(positionOrOffset);
        return this._blocks.find((block) => block.loc.start.offset <= offset && offset <= block.loc.end.offset);
    }
    documentAt(positionOrOffset) {
        const block = this.blockAt(positionOrOffset);
        return block ? this.getBlockDocument(block) : undefined;
    }
    parse() {
        const source = this.getText();
        const { descriptor, errors } = parse(source, {
            sourceMap: false,
            pad: 'space',
            filename: this.fsPath,
            shouldPadTemplate: true,
        });
        const prevIds = Object.keys(this.documents);
        this._descriptor = descriptor;
        this._parseErrors = errors;
        this._blocks = [descriptor.script, descriptor.template].filter(isNotNull);
        this._blocks = this._blocks
            .concat(descriptor.styles)
            .concat(descriptor.customBlocks);
        this._blocks.sort((a, b) => a.loc.start.offset - b.loc.start.offset);
        if (prevIds.length) {
            prevIds.forEach((id) => {
                const selector = querystring.parse(id);
                if (selector && 'index' in selector) {
                    const block = this._blocks[selector.index];
                    this.setBlockDocument(block, this.createBlockDocument(block));
                }
            });
        }
        if (descriptor.script) {
            this.setBlockDocument(descriptor.script, this.createBlockDocument(descriptor.script));
        }
        if (descriptor.template) {
            this.setBlockDocument(descriptor.template, this.createBlockDocument(descriptor.template));
        }
        return this;
    }
    createBlockDocument(block) {
        if (!block)
            return null;
        const selector = this.getSelectorFor(block);
        const id = querystring.stringify(selector);
        const uri = this.getUriFor(block);
        const languageId = getLanguageIdForBlock(block);
        const prevDocument = this.documents[id];
        if (prevDocument) {
            if (prevDocument.uri === uri && prevDocument.languageId === languageId) {
                return VirtualTextDocument.update(prevDocument, [{ text: block.content }], this.version);
            }
        }
        return VirtualTextDocument.create(this, uri, languageId, this.version, block.content, selector);
    }
    all() {
        return Object.values(this.documents).filter(isNotNull);
    }
    forTS() {
        return Object.values(this.documents)
            .filter(isNotNull)
            .filter((document) => /^(typescript|javascript)$/.test(document.languageId));
    }
    getBlockDocument(block) {
        if (!block)
            return;
        const selector = this.getSelectorFor(block);
        if (!selector)
            return;
        const id = querystring.stringify(selector);
        if ('index' in selector &&
            !/^(script|template)$/.test(selector.type) &&
            !this.documents[id]) {
            const block = this.blocks[selector.index];
            if (block)
                this.setBlockDocument(block, this.createBlockDocument(block));
        }
        return this.documents[id];
    }
    setBlockDocument(block, document) {
        const id = querystring.stringify(this.getSelectorFor(block));
        if (document) {
            this.documents[id] = document;
        }
        else {
            delete this.documents[id];
        }
    }
    getUriFor(block) {
        const selector = this.getSelectorFor(block);
        const uri = URI.file(this.uriObject.path +
            virtualFileNameSep +
            ('index' in selector
                ? `${selector.type}__${selector.index}`
                : selector.type) +
            '.' +
            this.getLanguageFor(block));
        return uri.toString().replace(/^file/, 'vue');
    }
    getSelectorFor(block) {
        if (!block)
            return { type: '', index: -1 };
        if (block === 'script') {
            return { type: 'script' };
        }
        else if (block === 'template') {
            return { type: 'template' };
        }
        else if (typeof block === 'string') {
            const virtual = parseVirtualFileUri(block);
            if (!virtual)
                throw new Error(`Malforated virtual document uri: ${block}`);
            return virtual.selector;
        }
        else if ('index' in block) {
            return block;
        }
        else {
            switch (block.type) {
                case 'script':
                    return { type: 'script' };
                case 'template':
                    return { type: 'template' };
                default:
                    return {
                        type: block.type,
                        index: this.blocks.indexOf(block),
                    };
            }
        }
    }
    getLanguageFor(block) {
        return (block.lang ||
            BlockLanguageMap[block.type] ||
            'txt');
    }
    static create(uri, languageId, version, content) {
        if (!isVueFile(uri))
            throw new Error(`Not a .vue file: ${uri}`);
        return new VueTextDocument(TextDocument.create(uri, languageId, version, content));
    }
    static update(document, changes, version) {
        document.internal = TextDocument.update(document.internal, changes, version);
        return document.parse();
    }
    static fsPathToUri(fsPath) {
        const uri = URI.file(fsPath).toString();
        if (isVirtualFile(fsPath)) {
            return uri.replace(/^[^:]+/, 'vue');
        }
        return uri;
    }
}
function parseVirtualFileUri(fileNameOrUri) {
    const uri = URI.parse(asUri(fileNameOrUri));
    if (uri.scheme === 'vue') {
        const [container, selector] = uri.fsPath.split(virtualFileNameSep);
        if (!selector.includes('.'))
            throw new Error('Malformated virtual file uri.');
        const [block, index] = selector
            .substr(0, selector.lastIndexOf('.'))
            .split('__');
        return {
            // TODO: DO NOT ASSUME FILE SCHEME!
            uri: URI.file(container).toString(),
            selector: (index
                ? {
                    type: block,
                    index: parseInt(index, 10),
                }
                : { type: block }),
        };
    }
    return null;
}
class DocumentStore {
    constructor(resolve, normalize = (uri) => uri) {
        this.resolve = resolve;
        this.normalize = normalize;
        this.map = new Map();
        this.reverseUriMap = new Map();
    }
    getNormalizedUri(uri) {
        return this.reverseUriMap.get(this.normalize(uri)) || uri;
    }
    has(uri) {
        return this.map.has(this.getNormalizedUri(uri));
    }
    get(uri) {
        return this.map.get(this.getNormalizedUri(uri)) || this.loadSync(uri);
    }
    set(uri, document) {
        this.map.set(uri, document);
        this.reverseUriMap.set(uri.toLowerCase(), uri);
    }
    delete(uri) {
        return this.map.delete(this.getNormalizedUri(uri));
    }
    all() {
        return Array.from(this.map.keys());
    }
    dispose() {
        this.map.clear();
    }
    loadSync(uri) {
        const document = this.resolve(uri);
        if (document) {
            this.set(uri, document);
        }
        return document;
    }
}
function isNotNull(value) {
    return value != null;
}
function getLanguageIdForBlock(block) {
    switch (block.type) {
        case 'template':
            return getLanguageIdFromExtension(block.lang) || 'vue-html';
        case 'script':
            return getLanguageIdFromExtension(block.lang) || 'javascript';
        case 'style':
            return getLanguageIdFromExtension(block.lang) || 'css';
        default:
            return getLanguageIdFromExtension(block.lang) || 'text';
    }
}
function getLanguageIdFromExtension(ext) {
    switch (ext) {
        case 'js':
            return 'javascript';
        case 'ts':
            return 'typescript';
    }
    return ext;
}

// Vue File: foo.vue
// Virtual Script File: foo.vue____script.js
// Virtual Template File: foo.vue____template.vue-html
// Virtual Render File: foo.vue____render.js
// Virtual Style File: foo.vue____style__0.css
const virtualFileNameSep$1 = '____';
function isVueFile$1(fileName) {
    return fileName.endsWith('.vue');
}
function isVirtualFile$1(fileName) {
    return fileName.indexOf('.vue' + virtualFileNameSep$1) > 0;
}
function asUri$1(fileNameOrUri) {
    if (/^[a-z-]+:\/\//i.test(fileNameOrUri))
        return fileNameOrUri;
    return VueTextDocument$1.fsPathToUri(fileNameOrUri);
}
class VirtualTextDocument$1 {
    constructor(internal, container, selector) {
        this.internal = internal;
        this.container = container;
        this.selector = selector;
        this.uriObject = URI.parse(this.internal.uri);
    }
    get version() {
        return this.internal.version;
    }
    get languageId() {
        return this.internal.languageId;
    }
    get lineCount() {
        return this.internal.lineCount;
    }
    get uri() {
        return this.internal.uri;
    }
    get fsPath() {
        return this.uriObject.fsPath;
    }
    get content() {
        return this.getText();
    }
    getText(range) {
        return this.internal.getText(range);
    }
    positionAt(offset) {
        return this.internal.positionAt(offset);
    }
    offsetAt(position) {
        return this.internal.offsetAt(position);
    }
    toJSON() {
        return {
            uri: this.uri,
            languageId: this.languageId,
            vertion: this.version,
            content: this.getText(),
        };
    }
    static create(container, uri, languageId, version, content, selector) {
        return new VirtualTextDocument$1(TextDocument.create(uri, languageId, version, content), container, selector);
    }
    static update(document, changes, version) {
        document.internal = TextDocument.update(document.internal, changes, version);
        return document;
    }
}
const BlockLanguageMap$1 = {
    script: 'js',
    template: 'vue-html',
    style: 'css',
    docs: 'md',
};
class VueTextDocument$1 {
    constructor(internal) {
        this.internal = internal;
        this._blocks = [];
        this._parseErrors = [];
        this.documents = {};
        this.uriObject = URI.parse(this.internal.uri);
        this.parse();
    }
    get version() {
        return this.internal.version;
    }
    get languageId() {
        return this.internal.languageId;
    }
    get lineCount() {
        return this.internal.lineCount;
    }
    get uri() {
        return this.internal.uri;
    }
    get fsPath() {
        return this.uriObject.fsPath;
    }
    get descriptor() {
        return this._descriptor;
    }
    get blocks() {
        return this._blocks;
    }
    get parseErrors() {
        return this._parseErrors;
    }
    get content() {
        return this.getText();
    }
    getText(range) {
        return this.internal.getText(range);
    }
    positionAt(offset) {
        return this.internal.positionAt(offset);
    }
    offsetAt(position) {
        return this.internal.offsetAt(position);
    }
    blockAt(positionOrOffset) {
        const offset = typeof positionOrOffset === 'number'
            ? positionOrOffset
            : this.offsetAt(positionOrOffset);
        return this._blocks.find((block) => block.loc.start.offset <= offset && offset <= block.loc.end.offset);
    }
    documentAt(positionOrOffset) {
        const block = this.blockAt(positionOrOffset);
        return block ? this.getBlockDocument(block) : undefined;
    }
    parse() {
        const source = this.getText();
        const { descriptor, errors } = parse(source, {
            sourceMap: false,
            pad: 'space',
            filename: this.fsPath,
            shouldPadTemplate: true,
        });
        const prevIds = Object.keys(this.documents);
        this._descriptor = descriptor;
        this._parseErrors = errors;
        this._blocks = [descriptor.script, descriptor.template].filter(isNotNull$1);
        this._blocks = this._blocks
            .concat(descriptor.styles)
            .concat(descriptor.customBlocks);
        this._blocks.sort((a, b) => a.loc.start.offset - b.loc.start.offset);
        if (prevIds.length) {
            prevIds.forEach((id) => {
                const selector = querystring.parse(id);
                if (selector && 'index' in selector) {
                    const block = this._blocks[selector.index];
                    this.setBlockDocument(block, this.createBlockDocument(block));
                }
            });
        }
        if (descriptor.script) {
            this.setBlockDocument(descriptor.script, this.createBlockDocument(descriptor.script));
        }
        if (descriptor.template) {
            this.setBlockDocument(descriptor.template, this.createBlockDocument(descriptor.template));
        }
        return this;
    }
    createBlockDocument(block) {
        if (!block)
            return null;
        const selector = this.getSelectorFor(block);
        const id = querystring.stringify(selector);
        const uri = this.getUriFor(block);
        const languageId = getLanguageIdForBlock$1(block);
        const prevDocument = this.documents[id];
        if (prevDocument) {
            if (prevDocument.uri === uri && prevDocument.languageId === languageId) {
                return VirtualTextDocument$1.update(prevDocument, [{ text: block.content }], this.version);
            }
        }
        return VirtualTextDocument$1.create(this, uri, languageId, this.version, block.content, selector);
    }
    all() {
        return Object.values(this.documents).filter(isNotNull$1);
    }
    forTS() {
        return Object.values(this.documents)
            .filter(isNotNull$1)
            .filter((document) => /^(typescript|javascript)$/.test(document.languageId));
    }
    getBlockDocument(block) {
        if (!block)
            return;
        const selector = this.getSelectorFor(block);
        if (!selector)
            return;
        const id = querystring.stringify(selector);
        if ('index' in selector &&
            !/^(script|template)$/.test(selector.type) &&
            !this.documents[id]) {
            const block = this.blocks[selector.index];
            if (block)
                this.setBlockDocument(block, this.createBlockDocument(block));
        }
        return this.documents[id];
    }
    setBlockDocument(block, document) {
        const id = querystring.stringify(this.getSelectorFor(block));
        if (document) {
            this.documents[id] = document;
        }
        else {
            delete this.documents[id];
        }
    }
    getUriFor(block) {
        const selector = this.getSelectorFor(block);
        const uri = URI.file(this.uriObject.path +
            virtualFileNameSep$1 +
            ('index' in selector
                ? `${selector.type}__${selector.index}`
                : selector.type) +
            '.' +
            this.getLanguageFor(block));
        return uri.toString().replace(/^file/, 'vue');
    }
    getSelectorFor(block) {
        if (!block)
            return { type: '', index: -1 };
        if (block === 'script') {
            return { type: 'script' };
        }
        else if (block === 'template') {
            return { type: 'template' };
        }
        else if (typeof block === 'string') {
            const virtual = parseVirtualFileUri$1(block);
            if (!virtual)
                throw new Error(`Malforated virtual document uri: ${block}`);
            return virtual.selector;
        }
        else if ('index' in block) {
            return block;
        }
        else {
            switch (block.type) {
                case 'script':
                    return { type: 'script' };
                case 'template':
                    return { type: 'template' };
                default:
                    return {
                        type: block.type,
                        index: this.blocks.indexOf(block),
                    };
            }
        }
    }
    getLanguageFor(block) {
        return (block.lang ||
            BlockLanguageMap$1[block.type] ||
            'txt');
    }
    static create(uri, languageId, version, content) {
        if (!isVueFile$1(uri))
            throw new Error(`Not a .vue file: ${uri}`);
        return new VueTextDocument$1(TextDocument.create(uri, languageId, version, content));
    }
    static update(document, changes, version) {
        document.internal = TextDocument.update(document.internal, changes, version);
        return document.parse();
    }
    static fsPathToUri(fsPath) {
        const uri = URI.file(fsPath).toString();
        if (isVirtualFile$1(fsPath)) {
            return uri.replace(/^[^:]+/, 'vue');
        }
        return uri;
    }
}
function parseVirtualFileUri$1(fileNameOrUri) {
    const uri = URI.parse(asUri$1(fileNameOrUri));
    if (uri.scheme === 'vue') {
        const [container, selector] = uri.fsPath.split(virtualFileNameSep$1);
        if (!selector.includes('.'))
            throw new Error('Malformated virtual file uri.');
        const [block, index] = selector
            .substr(0, selector.lastIndexOf('.'))
            .split('__');
        return {
            // TODO: DO NOT ASSUME FILE SCHEME!
            uri: URI.file(container).toString(),
            selector: (index
                ? {
                    type: block,
                    index: parseInt(index, 10),
                }
                : { type: block }),
        };
    }
    return null;
}
function isNotNull$1(value) {
    return value != null;
}
function getLanguageIdForBlock$1(block) {
    switch (block.type) {
        case 'template':
            return getLanguageIdFromExtension$1(block.lang) || 'vue-html';
        case 'script':
            return getLanguageIdFromExtension$1(block.lang) || 'javascript';
        case 'style':
            return getLanguageIdFromExtension$1(block.lang) || 'css';
        default:
            return getLanguageIdFromExtension$1(block.lang) || 'text';
    }
}
function getLanguageIdFromExtension$1(ext) {
    switch (ext) {
        case 'js':
            return 'javascript';
        case 'ts':
            return 'typescript';
    }
    return ext;
}

function removeVirtualSuffixFromFileName(fileName) {
    if (isVirtualFile$1(String(fileName))) {
        return fileName.substr(0, fileName.lastIndexOf(virtualFileNameSep$1));
    }
    return fileName;
}
const virtualFileRegex = new RegExp(`(?<=\\.vue)${virtualFileNameSep$1}(script|template|style|customBlock|render)(__[0-9]+)?(\\.[A-Za-z0-9_-]+)?`, 'g');
function removeVirtualSuffixFromText(text) {
    return text.replace(virtualFileRegex, '');
}
function prepareDocumentSpan(span) {
    span.fileName = removeVirtualSuffixFromFileName(span.fileName);
    if (span.originalFileName) {
        span.originalFileName = removeVirtualSuffixFromFileName(span.originalFileName);
    }
}

function prepareCodeFixAction(fileName, result) {
    result.forEach(item => {
        item.changes.forEach(item => {
            // TODO: Can be new file!!!
            item.fileName = removeVirtualSuffixFromFileName(fileName);
        });
    });
    return result;
}

function prepareDefinitionAndBoundSpan(fileName, result) {
    var _a;
    (_a = result.definitions) === null || _a === void 0 ? void 0 : _a.forEach(prepareDocumentSpan);
    return result;
}
function prepareDefinitionInfo(fileName, result) {
    result.forEach(prepareDocumentSpan);
    return result;
}
function prepareImplementationLocation(fileName, result) {
    result.forEach(prepareDocumentSpan);
    return result;
}
function prepareReferenceEntry(fileName, result) {
    result.forEach(prepareDocumentSpan);
    return result;
}
function prepareReferenceSymbol(fileName, result) {
    result.forEach((symbol) => {
        prepareDocumentSpan(symbol.definition);
        symbol.references.forEach(prepareDocumentSpan);
    });
    return result;
}

function prepareSyntacticDiagnostics(fileName, result) {
    result.forEach((item) => {
        if (typeof item.messageText === 'string') {
            item.messageText = removeVirtualSuffixFromText(item.messageText);
        }
    });
    return result;
}
function prepareSuggestionDiagnostics(fileName, result) {
    result.forEach((item) => {
        if (typeof item.messageText === 'string') {
            item.messageText = removeVirtualSuffixFromText(item.messageText);
        }
    });
    return result;
}
function prepareSemanticDiagnostics(fileName, result) {
    result.forEach((item) => {
        if (typeof item.messageText === 'string') {
            item.messageText = removeVirtualSuffixFromText(item.messageText);
        }
    });
    return result;
}

function prepareApplicableRefactorInfo(fileName, result) {
    return result;
}
function prepareRefactorEditInfo(fileName, result) {
    if (result.renameFilename) {
        result.renameFilename = removeVirtualSuffixFromFileName(result.renameFilename);
    }
    result.edits.forEach(edit => {
        edit.fileName = removeVirtualSuffixFromFileName(edit.fileName);
    });
    return result;
}

function prepareQuickInfo(fileName, result) {
    var _a;
    (_a = result.displayParts) === null || _a === void 0 ? void 0 : _a.forEach((part) => {
        part.text = removeVirtualSuffixFromText(part.text);
    });
    return result;
}

function prepareDocumentHighlights(fileName, result) {
    result.forEach((highlight) => {
        highlight.fileName = removeVirtualSuffixFromFileName(highlight.fileName);
        highlight.highlightSpans.forEach((span) => {
            if (span.fileName)
                span.fileName = removeVirtualSuffixFromFileName(span.fileName);
        });
    });
    return result;
}

function prepareClassifiedSpans(fileName, result) {
    return result;
}
function prepareClassifications(fileName, result) {
    return result;
}

function prepareCompletionsEntryDetail(fileName, result) {
    var _a;
    (_a = result.codeActions) === null || _a === void 0 ? void 0 : _a.forEach((codeAction) => {
        var _a;
        (_a = codeAction.changes) === null || _a === void 0 ? void 0 : _a.forEach((change) => {
            change.fileName = removeVirtualSuffixFromFileName(change.fileName);
        });
    });
    result.displayParts.forEach((displayPart) => (displayPart.text = removeVirtualSuffixFromText(displayPart.text)));
    return result;
}
function prepareSymbol(fileName, result) {
    return result;
}

function prepareSignatureHelpItems(fileName, result) {
    return result;
}

function prepareRenameInfo(fileName, result) {
    if ('fileToRename' in result && result.fileToRename) {
        result.fileToRename = removeVirtualSuffixFromFileName(result.fileToRename);
    }
    return result;
}
function prepareRenameLocation(fileName, result) {
    result.forEach((item) => {
        item.fileName = removeVirtualSuffixFromFileName(item.fileName);
        if (item.originalFileName) {
            item.originalFileName = removeVirtualSuffixFromFileName(item.originalFileName);
        }
    });
    return result;
}

function prepareCallHierarchy(fileName, result) {
    if (Array.isArray(result)) {
        result.forEach(item => {
            item.file = removeVirtualSuffixFromFileName(item.file);
        });
    }
    else {
        result.file = removeVirtualSuffixFromFileName(result.file);
    }
    return result;
}
function prepareCallHierachyIncomingCalls(fileName, result) {
    result.forEach(item => {
        item.from.file = removeVirtualSuffixFromFileName(item.from.file);
    });
    return result;
}
function prepareCallHierachyOutgoingCalls(fileName, result) {
    result.forEach(item => {
        item.to.file = removeVirtualSuffixFromFileName(item.to.file);
    });
    return result;
}

function isNumber(any) {
    return typeof any === 'number';
}
function flat(items) {
    if (typeof Array.prototype.flat === 'function') {
        return items.flat(Infinity);
    }
    const result = [];
    items.forEach((item) => {
        if (Array.isArray(item)) {
            result.push(...flat(item));
        }
        else {
            result.push(item);
        }
    });
    return result;
}
function firstOrDefault(items, defaultValue) {
    if (items.length)
        return items[0];
    return defaultValue();
}
function chain(value) {
    return {
        next(fn) {
            if (value == undefined) {
                return chain(undefined);
            }
            try {
                return chain(fn(value));
            }
            catch (error) {
                console.log('Error in chain: ' + error.message + ' ' + error.stack);
                return chain(undefined);
            }
        },
        // @ts-ignore
        end(defaultValue) {
            return value == undefined ? defaultValue : value;
        },
    };
}
class VueContext {
    constructor(documents, logger) {
        this.documents = documents;
        this.logger = logger;
    }
    getFileNameFromVirtualFileName(fileName) {
        return fileName.split(virtualFileNameSep$1).shift();
    }
    isVueFile(fileName) {
        return isVueFile$1(fileName);
    }
    isVirtualFile(fileName) {
        return isVirtualFile$1(fileName);
    }
    getVueDocument(fileNameOrUri) {
        return this.documents.get(asUri$1(fileNameOrUri));
    }
    getVirtualDocument(fileNameOrUri) {
        return chain(parseVirtualFileUri$1(fileNameOrUri))
            .next((ref) => { var _a; return ((_a = this.getVueDocument(ref.uri)) === null || _a === void 0 ? void 0 : _a.getBlockDocument(ref.selector)) || null; })
            .end(null);
    }
    getVirtualDocumentAt(fileName, position) {
        const document = this.getVueDocument(fileName);
        if (!document)
            return null;
        return document.getBlockDocument(document.blockAt(position)) || null;
    }
    getFileNameAt(fileName, position, isRenderFunctionAllowed = false) {
        var _a;
        if (this.isVueFile(fileName)) {
            // TODO: Only return script or render fileName
            const virtualFileName = (_a = this.getVirtualDocumentAt(fileName, position)) === null || _a === void 0 ? void 0 : _a.fsPath;
            "development" !== "production" &&
                this.logger.log(`getFileNameAt(${fileName}, ${position}) = ${virtualFileName}`);
            return virtualFileName;
        }
        return fileName;
    }
    getInnerFileNames(fileName, isRenderFunctionAllowed = false) {
        if (this.isVueFile(fileName)) {
            // TODO: Only return script and render fileName
            const document = this.documents.get(fileName);
            if (!document)
                return [];
            const innerFileNames = document.forTS().map((document) => document.fsPath);
            "development" !== "production" &&
                this.logger.log(`getInnerFileNames(${fileName}) = ${JSON.stringify(innerFileNames)}`);
            return innerFileNames;
        }
        "development" !== "production" && this.logger.log(`getInnerFileNames(${fileName}) = [${fileName}]`);
        return [fileName];
    }
}
class VueLanguageServer {
    constructor(context, service, typescript) {
        this.context = context;
        this.service = service;
        this.typescript = typescript;
    }
    getProgram() {
        const program = this.service.getProgram();
        "development" !== "production" && this.context.logger.log(`VLS.getProgram() = ${typeof program}`);
        return program;
    }
    getSyntacticDiagnostics(fileName) {
        "development" !== "production" &&
            this.context.logger.log('getSyntacticDiagnostics file=' + fileName);
        const diagnostics = flat(this.context
            .getInnerFileNames(fileName)
            .map(this.service.getSyntacticDiagnostics));
        return prepareSyntacticDiagnostics(fileName, diagnostics);
    }
    getSemanticDiagnostics(fileName) {
        "development" !== "production" &&
            this.context.logger.log('getSemanticDiagnostics file=' + fileName);
        const diagnostics = flat(this.context
            .getInnerFileNames(fileName)
            .map(this.service.getSemanticDiagnostics));
        return prepareSemanticDiagnostics(fileName, diagnostics);
    }
    getSuggestionDiagnostics(fileName) {
        "development" !== "production" &&
            this.context.logger.log('getSemanticDiagnostics file=' + fileName);
        const diagnostics = flat(this.context
            .getInnerFileNames(fileName)
            .map(this.service.getSuggestionDiagnostics));
        return prepareSuggestionDiagnostics(fileName, diagnostics);
    }
    getSyntacticClassifications(fileName, span) {
        return chain(this.context.getFileNameAt(fileName, span.start))
            .next((fileName) => this.service.getSyntacticClassifications(fileName, span))
            .next((classifications) => prepareClassifiedSpans(fileName, classifications))
            .end([]);
    }
    getSemanticClassifications(fileName, span) {
        return chain(this.context.getFileNameAt(fileName, span.start))
            .next((fileName) => this.service.getSemanticClassifications(fileName, span))
            .next((classifications) => prepareClassifiedSpans(fileName, classifications))
            .end([]);
    }
    getEncodedSyntacticClassifications(fileName, span) {
        return chain(this.context.getFileNameAt(fileName, span.start))
            .next((fileName) => this.service.getEncodedSyntacticClassifications(fileName, span))
            .next((classifications) => prepareClassifications(fileName, classifications))
            .end({ spans: [], endOfLineState: this.typescript.EndOfLineState.None });
    }
    getEncodedSemanticClassifications(fileName, span) {
        return chain(this.context.getFileNameAt(fileName, span.start))
            .next((fileName) => this.service.getEncodedSemanticClassifications(fileName, span))
            .next((classifications) => prepareClassifications(fileName, classifications))
            .end({ spans: [], endOfLineState: this.typescript.EndOfLineState.None });
    }
    getCompletionsAtPosition(fileName, position, options) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next((fileName) => this.service.getCompletionsAtPosition(fileName, position, options))
            .end();
    }
    getCompletionEntryDetails(fileName, position, name, formatOptions, source, preferences) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next((fileName) => this.service.getCompletionEntryDetails(fileName, position, name, formatOptions, source, preferences))
            .next((detail) => prepareCompletionsEntryDetail(fileName, detail))
            .end();
    }
    getCompletionEntrySymbol(fileName, position, name, source) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next((fileName) => this.getCompletionEntrySymbol(fileName, position, name, source))
            .next((symbol) => prepareSymbol(fileName, symbol))
            .end();
    }
    getQuickInfoAtPosition(fileName, position) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next((fileName) => this.service.getQuickInfoAtPosition(fileName, position))
            .next((quickInfo) => prepareQuickInfo(fileName, quickInfo))
            .end();
    }
    getNameOrDottedNameSpan(fileName, startPos, endPos) {
        return chain(this.context.getFileNameAt(fileName, startPos))
            .next((fileName) => this.service.getNameOrDottedNameSpan(fileName, startPos, endPos))
            .end();
    }
    getBreakpointStatementAtPosition(fileName, position) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next((fileName) => this.service.getBreakpointStatementAtPosition(fileName, position))
            .end();
    }
    getSignatureHelpItems(fileName, position, options) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next((fileName) => this.service.getSignatureHelpItems(fileName, position, options))
            .next((items) => prepareSignatureHelpItems(fileName, items))
            .end();
    }
    getRenameInfo(fileName, position, options) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next((fileName) => this.service.getRenameInfo(fileName, position, options))
            .next((info) => prepareRenameInfo(fileName, info))
            .end({ canRename: false, localizedErrorMessage: '' });
    }
    findRenameLocations(fileName, position, findInStrings, findInComments, providePrefixAndSuffixTextForRename) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next((fileName) => this.service.findRenameLocations(fileName, position, findInStrings, findInComments, providePrefixAndSuffixTextForRename))
            .next((locations) => prepareRenameLocation(fileName, locations))
            .end();
    }
    getSmartSelectionRange(fileName, position) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next((fileName) => this.service.getSmartSelectionRange(fileName, position))
            .end({ textSpan: { start: position, length: 0 } });
    }
    getDefinitionAtPosition(fileName, position) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next((fileName) => this.service.getDefinitionAtPosition(fileName, position))
            .next((result) => prepareDefinitionInfo(fileName, result))
            .end();
    }
    getDefinitionAndBoundSpan(fileName, position) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next((fileName) => this.service.getDefinitionAndBoundSpan(fileName, position))
            .next((result) => prepareDefinitionAndBoundSpan(fileName, result))
            .end();
    }
    getTypeDefinitionAtPosition(fileName, position) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next((fileName) => this.service.getTypeDefinitionAtPosition(fileName, position))
            .next((result) => prepareDefinitionInfo(fileName, result))
            .end();
    }
    getImplementationAtPosition(fileName, position) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next((fileName) => this.service.getImplementationAtPosition(fileName, position))
            .next((result) => prepareImplementationLocation(fileName, result))
            .end();
    }
    getReferencesAtPosition(fileName, position) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next((fileName) => this.service.getReferencesAtPosition(fileName, position))
            .next((result) => prepareReferenceEntry(fileName, result))
            .end();
    }
    findReferences(fileName, position) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next((fileName) => this.service.findReferences(fileName, position))
            .next((result) => prepareReferenceSymbol(fileName, result))
            .end();
    }
    getDocumentHighlights(fileName, position, filesToSearch) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next((fileName) => this.service.getDocumentHighlights(fileName, position, filesToSearch
            .map((fileName) => this.context.getInnerFileNames(fileName))
            .filter(Boolean)
            .flat()))
            .next((highlights) => prepareDocumentHighlights(fileName, highlights))
            .end();
    }
    getOccurrencesAtPosition(fileName, position) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next((fileName) => this.service.getOccurrencesAtPosition(fileName, position))
            .next((references) => prepareReferenceEntry(fileName, references))
            .end();
    }
    getNavigateToItems(searchValue, maxResultCount, fileName, excludeDtsFiles) {
        return this.service.getNavigateToItems(searchValue, maxResultCount, fileName, // TODO: Convert .vue file to script
        excludeDtsFiles);
    }
    getNavigationBarItems(fileName) {
        return flat(this.context
            .getInnerFileNames(fileName)
            .map((fileName) => this.service.getNavigationBarItems(fileName)));
    }
    getNavigationTree(fileName) {
        // TODO: Do we need to merge navigation tree from <script> and <template> blocks?
        return firstOrDefault(this.context
            .getInnerFileNames(fileName)
            .map((fileName) => this.service.getNavigationTree(fileName)), () => ({
            kind: this.typescript.ScriptElementKind.moduleElement,
            text: '<unknown>',
            kindModifiers: 'none',
            nameSpan: undefined,
            spans: [],
        }));
    }
    prepareCallHierarchy(fileName, position) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next((fileName) => this.service.prepareCallHierarchy(fileName, position))
            .next((result) => prepareCallHierarchy(fileName, result))
            .end();
    }
    provideCallHierarchyIncomingCalls(fileName, position) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next((fileName) => this.service.provideCallHierarchyIncomingCalls(fileName, position))
            .next((result) => prepareCallHierachyIncomingCalls(fileName, result))
            .end([]);
    }
    provideCallHierarchyOutgoingCalls(fileName, position) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next((fileName) => this.service.provideCallHierarchyOutgoingCalls(fileName, position))
            .next((result) => prepareCallHierachyOutgoingCalls(fileName, result))
            .end([]);
    }
    getOutliningSpans(fileName) {
        return flat(this.context
            .getInnerFileNames(fileName)
            .map((fileName) => this.service.getOutliningSpans(fileName)));
    }
    getTodoComments(fileName, descriptors) {
        return flat(this.context
            .getInnerFileNames(fileName)
            .map((fileName) => this.service.getTodoComments(fileName, descriptors)));
    }
    getBraceMatchingAtPosition(fileName, position) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next((fileName) => this.service.getBraceMatchingAtPosition(fileName, position))
            .end([]);
    }
    getIndentationAtPosition(fileName, position, options) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next((fileName) => this.service.getIndentationAtPosition(fileName, position, options))
            .end(0);
    }
    getFormattingEditsForRange(fileName, start, end, options) {
        return chain(this.context.getFileNameAt(fileName, start))
            .next((fileName) => this.service.getFormattingEditsForRange(fileName, start, end, options))
            .end([]);
    }
    getFormattingEditsForDocument(fileName, options) {
        return flat(this.context
            .getInnerFileNames(fileName)
            .map((fileName) => this.service.getFormattingEditsForDocument(fileName, options)));
    }
    getFormattingEditsAfterKeystroke(fileName, position, key, options) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next((fileName) => this.service.getFormattingEditsAfterKeystroke(fileName, position, key, options))
            .end([]);
    }
    getDocCommentTemplateAtPosition(fileName, position) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next((fileName) => this.service.getDocCommentTemplateAtPosition(fileName, position))
            .end();
    }
    isValidBraceCompletionAtPosition(fileName, position, openingBrace) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next((fileName) => this.service.isValidBraceCompletionAtPosition(fileName, position, openingBrace))
            .end(false);
    }
    getJsxClosingTagAtPosition(fileName, position) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next((fileName) => this.service.getJsxClosingTagAtPosition(fileName, position))
            .end();
    }
    getSpanOfEnclosingComment(fileName, position, onlyMultiLine) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next((fileName) => this.service.getSpanOfEnclosingComment(fileName, position, onlyMultiLine))
            .end();
    }
    getCodeFixesAtPosition(fileName, start, end, errorCodes, formatOptions, preferences) {
        return chain(this.context.getFileNameAt(fileName, start))
            .next((fileName) => this.service.getCodeFixesAtPosition(fileName, start, end, errorCodes, formatOptions, preferences))
            .next((fixes) => prepareCodeFixAction(fileName, fixes))
            .end([]);
    }
    getCombinedCodeFix(scope, fixId, formatOptions, preferences) {
        return this.service.getCombinedCodeFix(scope, fixId, formatOptions, preferences);
    }
    applyCodeActionCommand(action, formatSettings) {
        return this.service.applyCodeActionCommand(action, formatSettings);
    }
    getApplicableRefactors(fileName, positionOrRange, preferences) {
        return chain(this.context.getFileNameAt(fileName, isNumber(positionOrRange) ? positionOrRange : positionOrRange.pos))
            .next((fileName) => this.service.getApplicableRefactors(fileName, positionOrRange, preferences))
            .next((refactors) => prepareApplicableRefactorInfo(fileName, refactors))
            .end([]);
    }
    getEditsForRefactor(fileName, formatOptions, positionOrRange, refactorName, actionName, preferences) {
        return chain(this.context.getFileNameAt(fileName, isNumber(positionOrRange) ? positionOrRange : positionOrRange.pos))
            .next((fileName) => this.service.getEditsForRefactor(fileName, formatOptions, positionOrRange, refactorName, actionName, preferences))
            .next((refactor) => prepareRefactorEditInfo(fileName, refactor))
            .end();
    }
    organizeImports(scope, formatOptions, preferences) {
        return chain(scope)
            .next((scope) => {
            if (scope.type === 'file') {
                const fileName = this.context
                    .getInnerFileNames(scope.fileName)
                    .find(Boolean);
                if (fileName)
                    return { ...scope, fileName };
                return undefined; // abort
            }
            return scope;
        })
            .next((scope) => this.service.organizeImports(scope, formatOptions, preferences))
            .next((results) => {
            results.forEach(result => {
                if (this.context.isVirtualFile(result.fileName)) {
                    // TODO: Ensure offset padding for script blocks
                    result.fileName = removeVirtualSuffixFromFileName(result.fileName);
                }
            });
            return results;
        })
            .end([]);
    }
    getEditsForFileRename(oldFilePath, newFilePath, formatOptions, preferences) {
        // TODO: Handle rename.
        return this.service.getEditsForFileRename(oldFilePath, newFilePath, formatOptions, preferences);
    }
    static create(context, server, typescript) {
        const instance = new VueLanguageServer(context, server, typescript);
        const proxy = Object.create(null);
        const keys = Object.keys(server);
        for (const key of keys) {
            proxy[key] =
                key in instance
                    ? (...args) => {
                        "development" !== "production" && console.log(`start server.${key}(${typeof args[0] === 'string' ? args[0] : ''})`);
                        // @ts-ignore
                        const result = instance[key].apply(instance, args);
                        "development" !== "production" && console.log(`end server.${key}(${typeof args[0] === 'string' ? args[0] : ''})`);
                        return result;
                    }
                    : (...args) => {
                        // @ts-ignore
                        const result = server[key](...args);
                        return result;
                    };
        }
        return proxy;
    }
}

let store;
function init({ typescript: ts, }) {
    function create(info) {
        function log(...args) {
            if ("development" !== "production") {
                info.project.projectService.logger.info(`vue:: ` +
                    args
                        .join(' ')
                        .replace(new RegExp(__dirname.split('@vuedx').shift(), 'g'), '~/'));
            }
        }
        store =
            store ||
                new DocumentStore((uri) => {
                    const fileName = URI.parse(uri).fsPath;
                    const content = info.serverHost.readFile(fileName) || '';
                    "development" !== "production" && console.log(`DocumentStore.load(file=${fileName})`);
                    return VueTextDocument.create(uri, 'vue', 0, content);
                }, (uri) => {
                    return URI.file(info.project.projectService.toPath(URI.parse(uri).fsPath)).toString();
                });
        function verbose(message) {
            // return log(message)
        }
        "development" !== "production" && log(`Create ts server with vue plugin...`);
        if ("development" !== "production") {
            console.log = log;
        }
        const context = new VueContext(store, { log });
        // @ts-ignore
        if (!info.project.__VUE__) {
            // @ts-ignore
            info.project.__VUE__ = true;
            info.project.projectService.setHostConfiguration({
                extraFileExtensions: [
                    {
                        extension: 'vue',
                        isMixedContent: true,
                        scriptKind: ts.ScriptKind.Deferred,
                    },
                ],
            });
        }
        function forVirtualDocument(document, fn) {
            const scriptInfo = info.project.projectService.getScriptInfo(document.fsPath);
            if (scriptInfo)
                fn(scriptInfo);
        }
        function patchFileSystem(key) {
            override(info[key], 'fileExists', (fn) => (path) => {
                "development" !== "production" &&
                    (context.isVueFile(path) || context.isVirtualFile(path)) &&
                    verbose();
                return isVirtualFile(path) || fn(path);
            });
            override(info[key], 'readFile', (fn) => (path) => {
                "development" !== "production" &&
                    (context.isVueFile(path) || context.isVirtualFile(path)) &&
                    verbose();
                return isVirtualFile(path)
                    ? context.getVirtualDocument(path).getText()
                    : fn(path);
            });
            override(info[key], 'readDirectory', (fn) => (path, extensions, exclude, include, depth) => {
                if (extensions) {
                    extensions = ['vue'].concat(extensions);
                }
                return fn(path, extensions, include, exclude, depth);
            });
            override(info[key], 'watchFile', (fn) => (path, callback, pollingInterval, options) => {
                "development" !== "production" &&
                    (context.isVueFile(path) || context.isVirtualFile(path)) &&
                    verbose();
                if (isVirtualFile(path)) {
                    const realPath = context.getFileNameFromVirtualFileName(path);
                    return fn(realPath, (_, eventKind) => callback(path, eventKind), pollingInterval, options);
                }
                return fn(path, callback, pollingInterval, options);
            });
        }
        function patchModuleResolution(key) {
            const cache = ts.createModuleResolutionCache(info.project.getCurrentDirectory(), (fileName) => fileName);
            override(info[key], 'resolveModuleNames', (fn) => function (moduleNames, containingFile, reusedNames, redirectedReference, options) {
                const local = { ...options };
                if (!local.paths) {
                    local.paths = {};
                }
                const skipLogging = context.isVirtualFile(containingFile);
                if (context.isVirtualFile(containingFile)) {
                    containingFile = context.getFileNameFromVirtualFileName(containingFile);
                }
                const containingDir = Path.dirname(containingFile);
                const perFolderCache = cache.getOrCreateCacheForDirectory(containingDir);
                const forVueFiles = moduleNames.map((moduleName) => {
                    var _a;
                    if (isVueFile(moduleName) || moduleName in local.paths) {
                        if (perFolderCache.has(moduleName)) {
                            return perFolderCache.get(moduleName).resolvedModule;
                        }
                        const result = ts.resolveModuleName(moduleName, containingFile, options, {
                            ...info.serverHost,
                            fileExists(fileName) {
                                if (info.serverHost.fileExists(fileName))
                                    return true;
                                else if (fileName.endsWith('.vue.ts')) {
                                    return info.serverHost.fileExists(fileName.replace(/\.ts$/, ''));
                                }
                                return info.serverHost.fileExists(fileName);
                            },
                        }, cache, redirectedReference);
                        if ((_a = result.resolvedModule) === null || _a === void 0 ? void 0 : _a.resolvedFileName.endsWith('.vue.ts')) {
                            const fileName = result.resolvedModule.resolvedFileName.replace(/\.vue\.ts$/, '.vue');
                            const document = context
                                .getVueDocument(fileName)
                                .getBlockDocument('script');
                            const resolvedModule = {
                                isExternalLibraryImport: moduleName.includes('node_modules'),
                                resolvedFileName: document.fsPath,
                                extension: fileNameToExtension(document.fsPath),
                            };
                            perFolderCache.set(moduleName, { resolvedModule });
                            return resolvedModule;
                        }
                    }
                });
                const forAllFiles = fn(moduleNames, containingFile, reusedNames, redirectedReference, options);
                const result = moduleNames.map((_, index) => forVueFiles[index] || forAllFiles[index]);
                if ("development" !== "production" &&
                    !skipLogging &&
                    (forVueFiles.some(Boolean) || isVueFile(containingFile))) {
                    log(`${key}.resolveModuleNames in ${containingFile} = ${JSON.stringify(options)}`);
                    info.project.projectService.logger.startGroup();
                    moduleNames.forEach((moduleName, index) => {
                        var _a;
                        log(`  ${moduleName} => ${(_a = result[index]) === null || _a === void 0 ? void 0 : _a.resolvedFileName}`);
                    });
                    info.project.projectService.logger.endGroup();
                }
                return result;
            });
        }
        function patchScriptInformation(key) {
            override(info[key], 'getScriptFileNames', (fn) => () => {
                const result = fn();
                const vue = result.filter(isVueFile);
                const virtual = result.filter(isVirtualFile);
                virtual.forEach((file) => result.push(context.getFileNameFromVirtualFileName(file)));
                vue.forEach((fileName) => {
                    var _a;
                    (_a = context
                        .getVueDocument(fileName)) === null || _a === void 0 ? void 0 : _a.forTS().forEach((virtual) => result.push(virtual.fsPath));
                });
                "development" !== "production" &&
                    log(`${key}.getScriptFileNames() = ${JSON.stringify(result)}`);
                return result;
            });
            // Script Information
            override(info[key], 'getScriptKind', (fn) => (path) => {
                let kind;
                if (isVueFile(path)) {
                    kind = ts.ScriptKind.Deferred;
                }
                else if (isVirtualFile(path)) {
                    kind = languageIdToScriptKind(context.getVirtualDocument(path).languageId);
                }
                else {
                    kind = fn(path);
                }
                return kind;
            });
            override(info[key], 'getScriptVersion', (fn) => (path) => {
                let version;
                if (isVueFile(path)) {
                    version = fn(path);
                }
                else if (isVirtualFile(path)) {
                    version = fn(context.getFileNameFromVirtualFileName(path));
                }
                else {
                    version = fn(path);
                }
                return version;
            });
            override(info[key], 'getScriptSnapshot', (fn) => (path) => {
                if (isVirtualFile(path)) {
                    return ts.ScriptSnapshot.fromString(context.getVirtualDocument(path).getText());
                }
                if (isVueFile(path)) ;
                return fn(path);
            });
        }
        function patchProjectService() {
            override(info.project.projectService, 'getScriptInfoForPath', (fn) => {
                const skip = new Set();
                return (fileName) => {
                    let scriptInfo = fn(fileName);
                    if (context.isVueFile(fileName)) {
                        if (!scriptInfo) {
                            if (skip.has(fileName.toString()))
                                return scriptInfo;
                            skip.add(fileName.toString());
                            scriptInfo = info.project.projectService.getOrCreateScriptInfoForNormalizedPath(ts.server.toNormalizedPath(fileName), false);
                            skip.delete(fileName.toString());
                        }
                        patchScriptInfo(fileName, scriptInfo);
                    }
                    if (!scriptInfo && context.isVirtualFile(fileName)) {
                        info.project.projectService.getScriptInfoForPath(info.project.projectService.toPath(context.getFileNameFromVirtualFileName(fileName)));
                        const virtual = fn(fileName);
                        if (virtual) {
                            "development" !== "production" &&
                                log(`project.projectService.getScriptInfoForPath(fileName=${virtual.fileName})`);
                        }
                        return virtual;
                    }
                    return scriptInfo;
                };
            });
        }
        function patchScriptInfo(fileName, scriptInfo, fileContent) {
            if (!scriptInfo)
                return;
            if (!isVueFile(fileName))
                return;
            // @ts-ignore hidden property on patched script info.
            if (!scriptInfo.__vue__) {
                Object.defineProperty(scriptInfo, '__vue__', {
                    enumerable: false,
                    configurable: false,
                    value: true,
                });
                "development" !== "production" &&
                    log(`project.projectService.getScriptInfoForPath(fileName=${scriptInfo.fileName})`);
                const uri = URI.file(scriptInfo.fileName).toString();
                if (!fileContent) {
                    const snapshot = scriptInfo.getSnapshot();
                    fileContent = snapshot.getText(0, snapshot.getLength());
                }
                const document = store.has(uri)
                    ? store.get(uri)
                    : VueTextDocument.create(uri, 'vue', 0, fileContent);
                if (!store.has(uri)) {
                    "development" !== "production" &&
                        log(`document created while patching (fileName=${document.fsPath})`);
                    store.set(uri, document);
                }
                const { projectService } = info.project;
                const isOpen = scriptInfo.isScriptOpen();
                const project = scriptInfo.isOrphan()
                    ? projectService.getDefaultProjectForFile(scriptInfo.fileName, false)
                    : scriptInfo.getDefaultProject();
                store.set(uri, document);
                const related = document.forTS().map((virtual) => {
                    const normalizedPath = ts.server.toNormalizedPath(virtual.fsPath);
                    const path = projectService.toPath(normalizedPath);
                    const scriptKind = languageIdToScriptKind(virtual.languageId);
                    const scriptInfo = new ts.server.ScriptInfo(info.serverHost, normalizedPath, scriptKind, false, path, 
                    // @ts-ignore
                    projectService.filenameToScriptInfoVersion.get(path));
                    // @ts-ignore
                    projectService.filenameToScriptInfo.set(scriptInfo.path, scriptInfo);
                    // @ts-ignore
                    projectService.filenameToScriptInfoVersion.delete(scriptInfo.path);
                    "development" !== "production" &&
                        log(`project.projectService.getScriptInfoForPath(fileName=${scriptInfo.fileName})`);
                    if (project)
                        scriptInfo.attachToProject(project);
                    if (!isOpen) {
                        // @ts-ignore
                        projectService.watchClosedScriptInfo(scriptInfo);
                    }
                    else {
                        // @ts-ignore
                        projectService.stopWatchingScriptInfo(scriptInfo);
                        if (project) {
                            projectService.openClientFile(virtual.fsPath);
                            project.addRoot(scriptInfo, normalizedPath);
                        }
                        else {
                            info.project.addRoot(scriptInfo, normalizedPath);
                        }
                    }
                    return scriptInfo;
                });
                override(scriptInfo, 'editContent', (editContent) => (start, end, text) => {
                    "development" !== "production" &&
                        log(`ScriptInfo.editContent(fileName=${scriptInfo.fileName}, start=${start}, end=${end}, text=${text})`);
                    const range = {
                        start: document.positionAt(start),
                        end: document.positionAt(end),
                    };
                    editContent(start, end, text);
                    const version = getLastNumberFromVersion(scriptInfo.getLatestVersion());
                    VueTextDocument.update(document, [{ range, text }], version);
                    document.forTS().forEach((virtual) => forVirtualDocument(virtual, (currentScriptInfo) => {
                        currentScriptInfo.open(virtual.getText());
                        if (!scriptInfo.isScriptOpen()) {
                            currentScriptInfo.close();
                        }
                    }));
                });
                override(scriptInfo, 'attachToProject', (fn) => (project) => {
                    const ok = fn(project);
                    if (ok) {
                        related.forEach((scriptInfo) => scriptInfo.attachToProject(project));
                    }
                    return ok;
                });
                override(scriptInfo, 'detachFromProject', (fn) => (project) => {
                    fn(project);
                    related.forEach((scriptInfo) => scriptInfo.detachFromProject(project));
                });
                override(scriptInfo, 'open', (fn) => (newText) => {
                    fn(newText);
                    related.forEach((scriptInfo) => scriptInfo.open(context.getVirtualDocument(scriptInfo.fileName).getText()));
                });
                override(scriptInfo, 'close', (fn) => (fileExists) => {
                    fn(fileExists);
                    related.forEach((scriptInfo) => scriptInfo.close(!!context.getVirtualDocument(scriptInfo.fileName)));
                });
                override(scriptInfo, 'detachAllProjects', (fn) => () => {
                    fn();
                    related.forEach((scriptInfo) => scriptInfo.detachAllProjects());
                });
            }
        }
        patchFileSystem('serverHost');
        patchScriptInformation('languageServiceHost');
        patchModuleResolution('languageServiceHost');
        patchProjectService();
        return VueLanguageServer.create(context, info.languageService, ts);
    }
    function languageIdToScriptKind(languageId) {
        switch (languageId) {
            case 'typescript':
                return ts.ScriptKind.TS;
            case 'javascript':
                return ts.ScriptKind.JS;
            case 'vue':
                return ts.ScriptKind.Deferred;
        }
        return ts.ScriptKind.Unknown;
    }
    function fileNameToExtension(fileName) {
        const ext = fileName.substr(fileName.lastIndexOf('.'));
        switch (ext) {
            case '.ts':
                return ts.Extension.Ts;
            case '.js':
                return ts.Extension.Js;
            case '.tsx':
                return ts.Extension.Tsx;
            case '.jsx':
                return ts.Extension.Jsx;
            case '.json':
                return ts.Extension.Json;
            default:
                return ts.Extension.Js;
        }
    }
    return {
        create,
    };
}
function getLastNumberFromVersion(version) {
    const parts = version.split(/[^0-9]+/);
    const ver = parts.pop();
    return Number(ver);
}
function override(object, key, handler) {
    // @ts-ignore
    const original = object[key] ? object[key].bind(object) : object[key];
    if (!original)
        "development" !== "production" && console.log(`ERROR: cannot find ${key}`);
    const fn = handler(original);
    // @ts-ignore
    object[key] = fn;
}

module.exports = init;
//# sourceMappingURL=typescript-plugin-vue.js.map
